import opengate_core as g4
import opengate as gate
import sys
from box import Box
import numpy as np
import itk


class ARFActor(g4.GateARFActor, gate.ActorBase):
    """
    The ARF Actor is attached to a volume.
    Every time a particle enter, it considers the energy and the direction of the particle.
    It runs the neural network model to provide the probability of detection in all energy windows.

    Output is an (FIXME itk ?numpy ?) image that can be retrieved with self.output_image
    """

    type_name = "ARFActor"

    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        # required user info, default values
        # user_info.arf_detector = None
        user_info.batch_size = 2e5
        user_info.pth_filename = None
        user_info.image_size = [128, 128]
        mm = gate.g4_units("mm")
        user_info.image_spacing = [4.41806 * mm, 4.41806 * mm]
        user_info.distance_to_crystal = 75 * mm
        user_info.verbose_batch = False
        user_info.output = ""

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateARFActor.__init__(self, user_info.__dict__)
        # import module
        self.garf = gate.import_garf()
        if self.garf is None:
            print("Cannot run GANSource")
            sys.exit()
        # create the default detector
        # self.user_info.arf_detector = gate.ARFDetector(self.user_info)
        # prepare output
        self.user_info.output_image = None
        self.g4_actor = None
        self.pth_filename = user_info.pth_filename
        self.param = Box()
        self.nn = None
        self.model = None
        self.model_data = None
        self.batch_nb = 0
        self.detected_particles = 0

    def __str__(self):
        u = self.user_info
        s = f'ARFActor "{u.name}"'
        return s

    def initialize(self):
        gate.ActorBase.initialize(self)
        # self.user_info.arf_detector.initialize(self)
        self.ActorInitialize()
        self.SetARFFunction(self.apply)
        self.user_info.output_image = None

        # load the pth file
        self.nn, self.model = self.garf.load_nn(
            self.pth_filename, gpu="auto", verbose=False
        )
        p = self.param
        p.gpu_batch_size = int(float(self.user_info.batch_size))

        # size and spacing (2D)
        p.image_size = self.user_info.image_size
        p.image_spacing = self.user_info.image_spacing
        p.distance_to_crystal = self.user_info.distance_to_crystal
        self.model_data = self.nn["model_data"]

        # output image: nb of energy windows times nb of runs (for rotation)
        p.nb_ene = self.model_data["n_ene_win"]
        p.nb_runs = len(self.simulation.run_timing_intervals)
        # size and spacing in 3D
        p.image_size = [p.nb_ene, p.image_size[0], p.image_size[1]]
        p.image_spacing = [p.image_spacing[0], p.image_spacing[1], 1]
        # create output image as np array
        p.output_size = [p.nb_ene * p.nb_runs, p.image_size[1], p.image_size[2]]
        self.output_image = np.zeros(p.output_size, dtype=np.float64)
        # compute offset
        p.psize = [
            p.image_size[1] * p.image_spacing[0],
            p.image_size[2] * p.image_spacing[1],
        ]
        p.hsize = np.divide(p.psize, 2.0)
        p.offset = [p.image_spacing[0] / 2.0, p.image_spacing[1] / 2.0]

    def apply(self, actor):
        # get values from cpp side
        energy = np.array(actor.fEnergy)
        px = np.array(actor.fPositionX)
        py = np.array(actor.fPositionY)
        dx = np.array(actor.fDirectionX)
        dy = np.array(actor.fDirectionY)

        # convert direction in angles # FIXME or CPP side ?
        degree = gate.g4_units("degree")
        theta = np.arccos(dy) / degree
        phi = np.arccos(dx) / degree

        # update
        self.batch_nb += 1
        self.detected_particles += energy.shape[0]

        # build the data
        x = np.column_stack((px, py, theta, phi, energy))

        # apply the neural network
        if self.user_info.verbose_batch:
            print(f"Apply ARF neural network to {energy.shape[0]} samples")
        ax = x[:, 2:5]  # two angles and energy
        w = self.garf.nn_predict(self.model, self.nn["model_data"], ax)

        # positions
        p = self.param
        angles = x[:, 2:4]
        t = self.garf.compute_angle_offset(angles, p.distance_to_crystal)
        cx = x[:, 0:2]
        cx = cx + t
        coord = (cx + p.hsize - p.offset) / p.image_spacing[0:2]
        coord = np.around(coord).astype(int)
        v = coord[:, 0]
        u = coord[:, 1]
        u, v, w_pred = self.garf.remove_out_of_image_boundaries(u, v, w, p.image_size)

        # do nothing if there is no hit in the image
        if u.shape[0] != 0:
            temp = np.zeros(p.image_size, dtype=np.float64)
            temp = self.garf.image_from_coordinates(temp, u, v, w_pred)
            # add to previous, at the correct slice location
            # the slice is : current_ene_window + run_id * nb_ene_windows
            run_id = self.simulation.g4_RunManager.GetCurrentRun().GetRunID()
            s = p.nb_ene * run_id
            self.output_image[s : s + p.nb_ene] = (
                self.output_image[s : s + p.nb_ene] + temp
            )

    def EndSimulationAction(self):
        g4.GateARFActor.EndSimulationAction(self)
        # process the remaining elements in the batch
        self.apply(self)
        # convert to itk image
        self.output_image = itk.image_from_array(self.output_image)
        # set spacing and origin like HitsProjectionActor
        spacing = self.user_info.image_spacing
        spacing = np.array([spacing[0], spacing[1], 1])
        size = np.array(self.param.image_size)
        size[0] = self.param.image_size[2]
        size[2] = self.param.image_size[0]
        origin = -size / 2.0 * spacing + spacing / 2.0
        origin[2] = 0
        self.output_image.SetSpacing(spacing)
        self.output_image.SetOrigin(origin)
        # convert double to float
        InputImageType = itk.Image[itk.D, 3]
        OutputImageType = itk.Image[itk.F, 3]
        castImageFilter = itk.CastImageFilter[InputImageType, OutputImageType].New()
        castImageFilter.SetInput(self.output_image)
        castImageFilter.Update()
        self.output_image = castImageFilter.GetOutput()
        # write ?
        if self.user_info.output:
            itk.imwrite(
                self.output_image, gate.check_filename_type(self.user_info.output)
            )



import opengate as gate
import opengate_core as g4
from box import Box
from anytree import LevelOrderIter

cut_particle_names = {
    "gamma": "gamma",
    "electron": "e-",
    "positron": "e+",
    "proton": "proton",
}


class PhysicsManager:
    """
    Everything related to the physics (lists, cuts etc) should be here.
    """

    def __init__(self, simulation):
        # Keep a pointer to the current simulation
        self.simulation = simulation
        # user options
        self.user_info = gate.PhysicsUserInfo(self)
        # main g4 physic list
        # main g4 physic list
        self.g4_physic_list = None
        self.g4_decay = None
        self.g4_radioactive_decay = None
        self.g4_cuts_by_regions = []
        # default values
        self._default_parameters()

    def __del__(self):
        # not really clear but it seems that we should delete user_info here
        # if not seg fault (sometimes) at the end
        del self.user_info
        pass

    def __str__(self):
        s = f"{self.user_info.physics_list_name} Decay: {self.user_info.enable_decay}"
        return s

    def _default_parameters(self):
        ui = self.user_info
        # keep the name to be able to come back to default
        self.default_physic_list = "QGSP_BERT_EMV"
        ui.physics_list_name = self.default_physic_list
        ui.enable_decay = False
        ui.production_cuts.world = Box()
        ui.production_cuts.world.gamma = -1  # -1 means = will be the phys list default
        ui.production_cuts.world.proton = -1
        ui.production_cuts.world.electron = -1
        ui.production_cuts.world.positron = -1
        ui.production_cuts.world.propagate_to_daughters = True
        """
        FIXME Energy range not clear : does not work in mono-thread mode
        Ignored for the moment (keep them to None)
        """
        """
        eV = gate.g4_units('eV')
        keV = gate.g4_units('keV')
        GeV = gate.g4_units('GeV')
        ui.energy_range_min = 250 * eV
        ui.energy_range_max = 0.5 * GeV
        """
        ui.energy_range_min = None
        ui.energy_range_max = None
        ui.apply_cuts = True

    def initialize(self):
        self.initialize_physics_list()
        self.initialize_decay()
        self.initialize_em_options()
        # self.initialize_cuts()

    def dump_available_physics_lists(self):
        factory = g4.G4PhysListFactory()
        s = (
            f"Phys List:     {factory.AvailablePhysLists()}\n"
            f"Phys List EM:  {factory.AvailablePhysListsEM()}\n"
            f"Phys List add: {gate.available_additional_physics_lists}"
        )
        return s

    def dump_cuts(self):
        if self.simulation.is_initialized:
            return self.dump_cuts_initialized()
        s = ""
        for c in self.user_info.production_cuts:
            s += f"{c} : {self.user_info.production_cuts[c]}\n"
        return s

    def set_cut(self, volume_name, particle, value):
        cuts = self.user_info.production_cuts
        if particle == "all":
            cuts[volume_name]["gamma"] = value
            cuts[volume_name]["electron"] = value
            cuts[volume_name]["positron"] = value
            cuts[volume_name]["proton"] = value
            return
        cuts[volume_name][particle] = value

    def dump_cuts_initialized(self):
        s = ""
        if not self.user_info.apply_cuts:
            s = f"Apply cuts is FALSE: following cuts are NOT used.\n"
        lvs = g4.G4LogicalVolumeStore.GetInstance()
        for i in range(lvs.size()):
            lv = lvs.Get(i)
            region = lv.GetRegion()
            cuts = region.GetProductionCuts()
            s += f"{region.GetName()}: "
            for p in cut_particle_names:
                v = cuts.GetProductionCut(cut_particle_names[p])
                s += f"{p} {v} mm "
            s += "\n"
        # remove last line break
        return s[:-1]

    def initialize_physics_list(self):
        pl_name = self.user_info.physics_list_name
        # Select the Physic List: check if simple ones
        if pl_name.startswith("G4"):
            self.g4_physic_list = gate.create_modular_physics_list(pl_name)
        else:
            # If not, select the Physic List from the Factory
            factory = g4.G4PhysListFactory()
            if not factory.IsReferencePhysList(pl_name):
                s = (
                    f"Cannot find the physic list : {pl_name}\n"
                    f"Known list are : {factory.AvailablePhysLists()}\n"
                    f"With EM : {factory.AvailablePhysListsEM()}\n"
                    f"Default is {self.default_physic_list}\n"
                    f"Help : https://geant4-userdoc.web.cern.ch/UsersGuides/PhysicsListGuide/html/physicslistguide.html"
                )
                gate.fatal(s)
            self.g4_physic_list = factory.GetReferencePhysList(pl_name)

    def initialize_decay(self):
        """
        G4DecayPhysics - defines all particles and their decay processes
        G4RadioactiveDecayPhysics - defines radioactiveDecay for GenericIon
        """
        if not self.user_info.enable_decay:
            return
        # check if decay/radDecay already exist in the physics list
        # (keep p and pp in self to prevent destruction)
        self.g4_decay = self.g4_physic_list.GetPhysics("Decay")
        if not self.g4_decay:
            self.g4_decay = g4.G4DecayPhysics(1)
            self.g4_physic_list.RegisterPhysics(self.g4_decay)
        self.g4_radioactive_decay = self.g4_physic_list.GetPhysics("G4RadioactiveDecay")
        if not self.g4_radioactive_decay:
            self.g4_radioactive_decay = g4.G4RadioactiveDecayPhysics(1)
            self.g4_physic_list.RegisterPhysics(self.g4_radioactive_decay)

    def initialize_em_options(self):
        pass

    def initialize_cuts(self):
        # range
        if (
            self.user_info.energy_range_min is not None
            and self.user_info.energy_range_max is not None
        ):
            gate.warning(f"WARNING !!! SetEnergyRange only works in MT mode")
            pct = g4.G4ProductionCutsTable.GetProductionCutsTable()
            pct.SetEnergyRange(
                self.user_info.energy_range_min, self.user_info.energy_range_max
            )
        # inherit production cuts
        self.propagate_cuts_to_child()
        # global cuts
        self.user_info.g4_em_parameters.SetApplyCuts(self.user_info.apply_cuts)
        if not self.user_info.apply_cuts:
            s = f"No production cuts (apply_cuts is False)"
            gate.warning(s)
            gate.fatal(
                "Not implemented: currently it crashes when apply_cuts is False. To be continued ..."
            )
            return
        # production cuts by region
        for region in self.user_info.production_cuts:
            self.set_region_cut(region)

    def propagate_cuts_to_child(self):
        tree = self.simulation.volume_manager.volumes_tree
        pc = self.user_info.production_cuts
        # loop on the tree, level order
        for node in LevelOrderIter(tree[gate.__world_name__]):
            if not node.parent:
                # this is the world, do nothing
                continue
            # get the user cuts for this region
            if node.name in pc:
                cuts = pc[node.name]
            else:
                gate.fatal(f"Cannot find region {node.name} in the cuts list: {pc}")
            # get the cuts for the parent
            if node.parent.name in pc:
                pcuts = pc[node.parent.name]
            else:
                gate.fatal(
                    f"Cannot find parent region {node.parent.name} in the cuts list: {pc}"
                )
            for p in cut_particle_names:
                # set the cut for the current region like the parent
                # but only if it was not set by user
                if p not in cuts:
                    if p in pcuts:
                        cuts[p] = pcuts[p]
                    else:
                        # special case when the parent=world, with default cuts
                        cuts[p] = -1
                        pcuts[p] = -1

    def set_region_cut(self, region):
        # get the values for this region
        cuts_values = self.user_info.production_cuts[region]
        # special case for world region
        if region == gate.__world_name__:
            region = "DefaultRegionForTheWorld"
        rs = g4.G4RegionStore.GetInstance()
        reg = rs.GetRegion(region, True)
        if not reg:
            l = ""
            for i in range(rs.size()):
                l += f"{rs.Get(i).GetName()} "
            s = f'Cannot find the region name "{region}". Knowns regions are: {l}'
            gate.warning(s)
            return
        # set the cuts for the region
        cuts = None
        for p in cuts_values:
            if p not in cut_particle_names:
                s = f'Cuts error : I find "{p}" while cuts can only be set for: {cut_particle_names.keys()}'
                gate.fatal(s)
            if not cuts:
                cuts = g4.G4ProductionCuts()
            a = cut_particle_names[p]
            v = cuts_values[p]
            if type(v) != int and type(v) != float:
                gate.fatal(
                    f"The cut value must be a number, while it is {v} in {cuts_values}"
                )
            if v < 0:
                v = self.g4_physic_list.GetDefaultCutValue()
            cuts.SetProductionCut(v, a)
        reg.SetProductionCuts(cuts)
        # keep the cut object to prevent deletion
        self.g4_cuts_by_regions.append(cuts)



from .GenericSource import *
from .GANSourceDefaultGenerator import GANSourceDefaultGenerator


class GANSource(GenericSource):
    """
    GAN source: the Generator produces particles
    Input is a neural network Generator trained with a GAN
    """

    type_name = "GAN"

    @staticmethod
    def set_default_user_info(user_info):
        GenericSource.set_default_user_info(user_info)
        # additional param
        user_info.pth_filename = None
        user_info.position_keys = None
        user_info.backward_distance = None
        user_info.direction_keys = None
        user_info.energy_key = None
        user_info.energy_threshold = -1
        user_info.weight_key = None
        user_info.time_key = None
        user_info.time_relative = True
        user_info.batch_size = 10000
        user_info.generator = None
        user_info.verbose_generator = False
        user_info.is_paired = False
        user_info.use_time = False
        user_info.use_weight = False
        # specific to conditional GAN
        user_info.cond_image = None
        user_info.compute_directions = False
        user_info.cond_debug = False

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GateGANSource()

    def __init__(self, user_info):
        super().__init__(user_info)

    def initialize(self, run_timing_intervals):
        # FIXME -> check input user_info

        # initialize the mother class generic source
        GenericSource.initialize(self, run_timing_intervals)

        # default generator or set by the user
        if self.user_info.generator is None:
            if self.user_info.cond_image is not None:
                voxelized_cond_generator = gate.VoxelizedSourceConditionGenerator(
                    self.user_info.cond_image, self
                )
                voxelized_cond_generator.compute_directions = (
                    self.user_info.compute_directions
                )
                gen = gate.GANSourceConditionalGenerator(
                    self.user_info,
                    voxelized_cond_generator.generate_condition,
                )
                self.user_info.generator = gen
            else:
                self.user_info.generator = GANSourceDefaultGenerator(self.user_info)

        # initialize the generator
        self.user_info.generator.initialize()

        # set the function pointer to the cpp side
        self.g4_source.SetGeneratorFunction(self.user_info.generator.generator)

        # weight ?
        if self.user_info.weight_key is None:
            self.g4_source.fUseWeight = False
        else:
            self.g4_source.fUseWeight = True

        # time ?
        if self.user_info.time_key is None:
            self.g4_source.fUseTime = False
        else:
            self.g4_source.fUseTime = True
            self.g4_source.fUseTimeRelative = self.user_info.time_relative



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import time
import itk
import numpy as np

paths = gate.get_default_test_paths(__file__)

n = int(1e5)

# read img
f = paths.data / "source_three_areas_crop_3.5mm.mhd"
img = itk.imread(str(f))


def test_voxelized(img, version):
    # create voxelized sampling
    v = gate.VoxelizedSourcePDFSampler(img, version=version)
    start = time.time()
    if version == 1:
        i, j, k = v.sample_indices(n)
    else:
        i, j, k = v.sample_indices_slower(n)
    end = time.time()
    print(f"Version {version}: done in {end - start:0.3f} sec")

    # create output image
    imga = itk.array_view_from_image(img)
    outa = np.zeros_like(imga)
    for a, b, c in zip(i, j, k):
        outa[a, b, c] += 1
    out = itk.image_from_array(outa)
    out.CopyInformation(img)
    # itk.imwrite(out, f"test047_vox_{version}.mhd")

    # test : compare with initial image
    nn = outa[imga != 0].sum()
    zz = outa[imga == 0].sum()
    is_ok = nn == n and zz == 0

    if not is_ok:
        gate.print_test(is_ok, f"Version {version}: ERROR ! {nn}/{n} and {zz}/0")
    else:
        gate.print_test(is_ok, f"Version {version}: test OK")

    return is_ok


is_ok = test_voxelized(img, 1)
is_ok = test_voxelized(img, 2) and is_ok
is_ok = test_voxelized(img, 3) and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.helpers import *

paths = gate.get_default_test_paths(__file__, "")

p = Box()
p.phantom_type = "analytic"
p.source_type = "analytic"
p.use_pet = True
p.use_gaga = False
p.a = 1e2

# debug
skip = False
# the seed is only for the G4 part, the GAN use his own seed
# so the results are not reproducible (yet)
seed = 812365478


def run(param):
    print("run ", param)
    cmd_line = (
        f"{paths.current}/test045_speedup.py -o AUTO --seed {seed} -p {param.phantom_type} "
        f"-s {param.source_type} -r Ga68 -a {param.a}"
    )
    if param.use_pet:
        cmd_line += " --pet "
    if param.use_gaga:
        cmd_line += " --gaga "
    out = f"test045_speedup_p_{param.phantom_type}_s_{param.source_type}_pet_{param.use_pet}_gaga_{param.use_gaga}.txt"
    if not skip:
        print("cmd line", cmd_line)
        r = os.system(f"python {cmd_line}")

    print("Output ", out)
    return out


# output
output = []

# Test 1
p.phantom_type = "analytic"
p.source_type = "analytic"
p.use_gaga = False
p.a = 1e3
out = run(p)
output.append(out)

# Test 2
p.phantom_type = "analytic"
p.source_type = "vox"
p.use_gaga = False
p.a = 1e3
out = run(p)
output.append(out)

# Test 3
p.phantom_type = "vox"
p.source_type = "vox"
p.use_gaga = False
p.a = 1e3
out = run(p)
output.append(out)

# Test 4
p.phantom_type = "analytic"
p.source_type = "analytic"
p.use_gaga = True
p.a = 1e3
out = run(p)
output.append(out)

# Test 5
p.phantom_type = "analytic"
p.source_type = "vox"
p.use_gaga = True
p.a = 1e3
out = run(p)
output.append(out)

# Test 6
p.phantom_type = "vox"
p.source_type = "vox"
p.use_gaga = True
p.a = 1e3
out = run(p)
output.append(out)

print(output)

# tests stats file
is_ok = True
for o in output:
    stats = gate.read_stat_file(paths.output / o)
    stats_ref = gate.read_stat_file(paths.output_ref / o)
    ok = gate.assert_stats(stats, stats_ref, 0.03)
    gate.print_test(ok, f"Check {o}")
    is_ok = is_ok and ok
    print()

# tests pet files
keys = [
    "GlobalTime",
    "PostPosition_X",
    "PostPosition_Y",
    "PostPosition_Z",
    "TotalEnergyDeposit",
    "TrackVolumeCopyNo",
]
tols = [10.0] * len(keys)
tols[keys.index("GlobalTime")] = 0.04
tols[keys.index("PostPosition_X")] = 7
tols[keys.index("PostPosition_Y")] = 8
tols[keys.index("PostPosition_Z")] = 1.3
tols[keys.index("TotalEnergyDeposit")] = 0.03
tols[keys.index("TrackVolumeCopyNo")] = 4.1
scalings = [1.0] * len(keys)
scalings[keys.index("GlobalTime")] = 1e-9  # time in ns
for o in output:
    o = o.replace(".txt", ".root")
    o1 = paths.output / o
    o2 = paths.output_ref / o
    img = paths.output / o.replace(".root", ".png")
    ok = gate.compare_root3(
        o1,
        o2,
        "Singles",
        "Singles",
        keys,
        keys,
        tols,
        scalings,
        scalings,
        img,
        hits_tol=5,
    )
    gate.print_test(ok, f"Check {o}")
    is_ok = is_ok and ok
    print()

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"

# set the world size like in the Gate macro
m = gate.g4_units("m")
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]

# add a simple waterbox volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

# default source for tests
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
source = sim.add_source("Generic", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 200000 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")
sim.add_actor("TestActor", "Stats2")

# create G4 objects
sim.initialize()

# start simulation
sim.start()

stats = sim.get_actor("Stats")
print(stats)
print("-" * 50)

stats = sim.get_actor("Stats2")
print(stats)

# FIXME todo
gate.test_ok(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect


def create_spect_simu(sim, paths, number_of_threads=1, activity_kBq=300):
    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = number_of_threads
    ui.check_volumes_overlap = False

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    kBq = 1000 * Bq

    # world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect = gate_spect.add_ge_nm67_spect_head(
        sim, "spect", collimator_type="lehr", debug=False
    )

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [15 * cm, 15 * cm, 15 * cm]
    waterbox.material = "G4_WATER"
    waterbox.translation = [0, 0, 0]
    blue = [0, 1, 1, 1]
    waterbox.color = blue

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 10 * mm
    cuts.world.electron = 10 * mm
    cuts.world.positron = 10 * mm
    cuts.world.proton = 10 * mm
    cuts.spect.gamma = 0.1 * mm
    cuts.spect.electron = 0.1 * mm
    cuts.spect.positron = 0.1 * mm

    # default source for tests
    # activity = 300 * kBq
    activity = activity_kBq * kBq
    beam1 = sim.add_source("Generic", "beam1")
    beam1.mother = waterbox.name
    beam1.particle = "gamma"
    beam1.energy.mono = 140.5 * keV
    beam1.position.type = "sphere"
    beam1.position.radius = 1 * cm
    beam1.position.translation = [0, 0, 0]
    beam1.direction.type = "momentum"
    beam1.direction.momentum = [0, 0, -1]
    beam1.direction.type = "iso"
    beam1.direction.acceptance_angle.volumes = ["spect"]
    beam1.direction.acceptance_angle.intersection_flag = True
    beam1.activity = activity / ui.number_of_threads

    beam2 = sim.add_source("Generic", "beam2")
    beam2.mother = waterbox.name
    beam2.particle = "gamma"
    beam2.energy.mono = 140.5 * keV
    beam2.position.type = "sphere"
    beam2.position.radius = 3 * cm
    beam2.position.translation = [18 * cm, 0, 0]
    # beam2.direction.type = 'momentum'
    beam2.direction.type = "iso"
    beam2.direction.acceptance_angle.volumes = ["spect"]
    beam2.direction.acceptance_angle.intersection_flag = True
    beam2.activity = activity / ui.number_of_threads

    beam3 = sim.add_source("Generic", "beam3")
    beam3.mother = waterbox.name
    beam3.particle = "gamma"
    beam3.energy.mono = 140.5 * keV
    beam3.position.type = "sphere"
    beam3.position.radius = 1 * cm
    beam3.position.translation = [0, 10 * cm, 0]
    # beam3.direction.type = 'momentum'
    beam3.direction.type = "iso"
    beam3.direction.acceptance_angle.volumes = ["spect"]
    beam3.direction.acceptance_angle.intersection_flag = True
    beam3.activity = activity / ui.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("HitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    hc.output = ""  # No output paths.output / 'test028.root'
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("HitsAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_hits_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    # sc.policy = 'EnergyWeightedCentroidPosition'
    sc.skip_attributes = ["KineticEnergy", "ProcessDefinedStep", "KineticEnergy"]
    sc.output = hc.output

    # EnergyWindows
    cc = sim.add_actor("HitsEnergyWindowsActor", "EnergyWindows")
    cc.mother = crystal.name
    cc.input_hits_collection = "Singles"
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
        # {'name': 'spectrum', 'min': 0 * keV, 'max': 5000 * keV}  # should be strictly equal to 'Singles'
    ]
    cc.output = hc.output

    # sec = gate.g4_units('second')
    # sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1 * sec]]

    # projection
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    # 2D binning projection
    proj = sim.add_actor("HitsProjectionActor", "Projection")
    proj.mother = crystal.name
    # we set two times the spectrum channel to compare with Gate output
    proj.input_hits_collections = ["Singles", "scatter", "peak140", "Singles"]
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.size = [128, 128]
    # proj.plane = 'XY' # not implemented yet
    proj.output = paths.output / "proj028_colli.mhd"

    return spect, proj



from .SimulationStatisticsActor import *
from .DoseActor import *
from .SourceInfoActor import *
from .PhaseSpaceActor import *
from .TestActor import *
from .HitsCollectionActor import *
from .HitsAdderActor import *
from .HitsEnergyWindowsActor import *
from .HitsProjectionActor import *
from .MotionVolumeActor import *
from .ARFActor import *
from .ARFTrainingDatasetActor import *

actor_type_names = {
    SimulationStatisticsActor,
    DoseActor,
    SourceInfoActor,
    PhaseSpaceActor,
    HitsCollectionActor,
    HitsAdderActor,
    HitsEnergyWindowsActor,
    HitsProjectionActor,
    MotionVolumeActor,
    ARFActor,
    ARFTrainingDatasetActor,
    TestActor,
}
actor_builders = gate.make_builders(actor_type_names)



import opengate_core as g4


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class UIsessionVerbose(g4.G4UIsession):
    """
    Print Geant4 Verbose with color
    """

    def __del__(self):
        pass

    def ReceiveG4cout(self, coutString):
        print(f"{bcolors.OKBLUE}{coutString}", end="")
        return 0

    def ReceiveG4cerr(self, cerrString):
        print(f"{bcolors.WARNING}{cerrString}", end="")
        return 0



import opengate as gate
import opengate_core as g4
from box import Box


class SourceManager:
    """
    Manage all the sources in the simulation.
    The function prepare_generate_primaries will be called during
    the main run loop to set the current time and source.
    """

    # G4RunManager::BeamOn takes an int as input. The max cpp int value is currently 2147483647
    # Python manage int differently (no limit), so we need to set the max value here.
    max_int = 2147483647

    def __init__(self, simulation):
        # Keep a pointer to the current simulation
        self.simulation = simulation
        # List of run times intervals
        self.run_timing_intervals = None
        self.current_run_interval = None
        # List of sources user info
        self.user_info_sources = {}
        # List of sources (GateSource), for all threads
        self.sources = []
        # The source manager will be constructed at build (during ActionManager)
        # Its task is to call GeneratePrimaries and loop over the sources
        # For MT, the master_source_manager is the MasterThread
        # The g4_thread_source_managers list all master source for all threads
        self.g4_master_source_manager = None
        self.g4_thread_source_managers = []
        # internal variables
        self.particle_table = None
        # Options dict for cpp SourceManager
        # will be set in create_g4_source_manager
        self.source_manager_options = Box()

    def __str__(self):
        """
        str only dump the user info on a single line
        """
        v = [v.name for v in self.user_info_sources.values()]
        s = f'{" ".join(v)} ({len(self.user_info_sources)})'
        return s

    def __del__(self):
        pass

    def dump(self):
        n = len(self.user_info_sources)
        s = f"Number of sources: {n}"
        for source in self.user_info_sources.values():
            a = f"\n {source}"
            s += gate.indent(2, a)
        return s

    def get_source_info(self, name):
        if name not in self.user_info_sources:
            gate.fatal(
                f"The source {name} is not in the current "
                f"list of sources: {self.user_info_sources}"
            )
        return self.user_info_sources[name]

    def get_source(self, name):
        n = len(self.g4_thread_source_managers)
        if n > 0:
            gate.warning(f"Cannot get source in multithread mode, use get_source_MT")
            return None
        for source in self.sources:
            if source.user_info.name == name:
                return source.g4_source
        gate.fatal(
            f'The source "{name}" is not in the current '
            f"list of sources: {self.user_info_sources}"
        )

    def get_source_MT(self, name, thread):
        n = len(self.g4_thread_source_managers)
        if n == 0:
            gate.warning(f"Cannot get source in mono-thread mode, use get_source")
            return None
        i = 0
        for source in self.sources:
            if source.user_info.name == name:
                if i == thread:
                    return source.g4_source
                i += 1
        gate.fatal(
            f'The source "{name}" is not in the current '
            f"list of sources: {self.user_info_sources}"
        )

    def add_source(self, source_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_sources, name)
        # init the user info
        s = gate.UserInfo("Source", source_type, name)
        # append to the list
        self.user_info_sources[name] = s
        # return the info
        return s

    def initialize(self, run_timing_intervals):
        self.run_timing_intervals = run_timing_intervals
        gate.assert_run_timing(self.run_timing_intervals)
        if len(self.user_info_sources) == 0:
            gate.warning(f"No source: no particle will be generated")

    def initialize_actors(self, actors):
        actors = [actors[a] for a in actors]
        self.g4_master_source_manager.SetActors(actors)

    def build(self):
        # create particles table # FIXME in physics ??
        self.particle_table = g4.G4ParticleTable.GetParticleTable()
        self.particle_table.CreateAllParticles()
        # create the master source for the masterThread
        self.g4_master_source_manager = self.create_g4_source_manager(False)
        return self.g4_master_source_manager

    def create_g4_source_manager(self, append=True):
        # -----------------------------
        # This is called by all threads
        # -----------------------------
        # This object is needed here, because it can only be
        # created after physics initialization
        ms = g4.GateSourceManager()
        # create all sources for this source manager (for all threads)
        for vu in self.user_info_sources.values():
            source = gate.new_element(vu, self.simulation)
            ms.AddSource(source.g4_source)
            source.initialize(self.run_timing_intervals)
            self.sources.append(source)
        # taking __dict__ allow to consider the class SimulationUserInfo as a dict
        sui = self.simulation.user_info.__dict__
        # warning: copy the simple elements from this dict (containing visu or verbose)
        for s in sui:
            if "visu" in s or "verbose_" in s:
                self.source_manager_options[s] = sui[s]
        ms.Initialize(self.run_timing_intervals, self.source_manager_options)
        # keep pointer to avoid deletion
        if append:
            self.g4_thread_source_managers.append(ms)
        return ms

    def start(self):
        # FIXME (1) later : may replace BeamOn with DoEventLoop
        # to allow better control on geometry between the different runs
        # (2) : check estimated nb of particle, warning if too large
        # start the master thread (only main thread)
        self.g4_master_source_manager.StartMasterThread()



import numpy as np

import opengate as gate
import opengate_core as g4
from box import Box
from scipy.spatial.transform import Rotation


class GenericSource(gate.SourceBase):
    """
    GeneriSource close to the G4 SPS, but a bit simpler.
    """

    type_name = "Generic"

    @staticmethod
    def set_default_user_info(user_info):
        gate.SourceBase.set_default_user_info(user_info)
        # initial user info
        user_info.particle = "gamma"
        user_info.ion = Box()
        user_info.n = 0
        user_info.activity = 0
        user_info.weight = -1
        user_info.weight_sigma = -1
        user_info.half_life = -1  # negative value is no half_life
        # ion
        user_info.ion = Box()
        user_info.ion.Z = 0  # Z: Atomic Number
        user_info.ion.A = 0  # A: Atomic Mass (nn + np +nlambda)
        user_info.ion.E = 0  # E: Excitation energy (i.e. for metastable)
        # position
        user_info.position = Box()
        user_info.position.type = "point"
        user_info.position.radius = 0
        user_info.position.sigma_x = 0
        user_info.position.sigma_y = 0
        user_info.position.size = [0, 0, 0]
        user_info.position.translation = [0, 0, 0]
        user_info.position.rotation = Rotation.identity().as_matrix()
        user_info.position.confine = None
        # angle (direction)
        user_info.direction = Box()
        user_info.direction.type = "iso"
        user_info.direction.momentum = [0, 0, 1]
        user_info.direction.focus_point = [0, 0, 0]
        user_info.direction.sigma = [0, 0]
        user_info.direction.acceptance_angle = Box()
        user_info.direction.acceptance_angle.volumes = []
        user_info.direction.acceptance_angle.intersection_flag = False
        user_info.direction.acceptance_angle.normal_flag = False
        user_info.direction.acceptance_angle.normal_vector = [0, 0, 1]
        deg = gate.g4_units("deg")
        user_info.direction.acceptance_angle.normal_tolerance = 3 * deg
        # energy
        user_info.energy = Box()
        user_info.energy.type = "mono"
        user_info.energy.mono = 0
        user_info.energy.sigma_gauss = 0
        user_info.energy.is_cdf = False
        user_info.energy.min_energy = None
        user_info.energy.max_energy = None

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GateGenericSource()

    def __init__(self, user_info):
        super().__init__(user_info)
        if not self.user_info.particle.startswith("ion"):
            return
        words = self.user_info.particle.split(" ")
        if len(words) > 1:
            self.user_info.ion.Z = words[1]
        if len(words) > 2:
            self.user_info.ion.A = words[2]
        if len(words) > 3:
            self.user_info.ion.E = words[3]

    def initialize(self, run_timing_intervals):
        # Check user_info type
        # if not isinstance(self.user_info, Box):
        #    gate.fatal(f'Generic Source: user_info must be a Box, but is: {self.user_info}')
        if not isinstance(self.user_info, gate.UserInfo):
            gate.fatal(
                f"Generic Source: user_info must be a UserInfo, but is: {self.user_info}"
            )
        if not isinstance(self.user_info.position, Box):
            gate.fatal(
                f"Generic Source: user_info.position must be a Box, but is: {self.user_info.position}"
            )
        if not isinstance(self.user_info.direction, Box):
            gate.fatal(
                f"Generic Source: user_info.direction must be a Box, but is: {self.user_info.direction}"
            )
        if not isinstance(self.user_info.energy, Box):
            gate.fatal(
                f"Generic Source: user_info.energy must be a Box, but is: {self.user_info.energy}"
            )

        # check energy type
        l = [
            "mono",
            "gauss",
            "F18_analytic",
            "O15_analytic",
            "C11_analytic",
            "spectrum",
            "range",
        ]
        l.extend(gate.all_beta_plus_radionuclides)
        if not self.user_info.energy.type in l:
            gate.fatal(
                f"Cannot find the energy type {self.user_info.energy.type} for the source {self.user_info.name}.\n"
                f"Available types are {l}"
            )

        # special case for beta plus energy spectra
        if self.user_info.particle == "e+":
            if self.user_info.energy.type in gate.all_beta_plus_radionuclides:
                data = gate.read_beta_plus_spectra(self.user_info.energy.type)
                ene = data[:, 0] / 1000  # convert from KeV to MeV
                proba = data[:, 1]
                cdf, total = gate.compute_cdf_and_total_yield(proba, ene)
                # total = total * 1000  # (because was in MeV)
                # self.user_info.activity *= total
                self.user_info.energy.is_cdf = True
                self.g4_source.SetEnergyCDF(ene)
                self.g4_source.SetProbabilityCDF(cdf)

        # initialize
        gate.SourceBase.initialize(self, run_timing_intervals)

        if self.user_info.n > 0 and self.user_info.activity > 0:
            gate.fatal(f"Cannot use both n and activity, choose one: {self.user_info}")
        if self.user_info.n == 0 and self.user_info.activity == 0:
            gate.fatal(f"Choose either n or activity : {self.user_info}")
        if self.user_info.activity > 0:
            self.user_info.n = -1
        if self.user_info.n > 0:
            self.user_info.activity = -1
        # warning for non used ?
        # check confine
        if self.user_info.position.confine:
            if self.user_info.position.type == "point":
                gate.warning(
                    f"In source {self.user_info.name}, "
                    f"confine is used, while position.type is point ... really ?"
                )


def get_source_skipped_particles(sim, source_name):
    ui = sim.user_info
    n = 0
    if ui.number_of_threads > 1 or ui.force_multithread_mode:
        for i in range(1, sim.user_info.number_of_threads + 1):
            s = sim.get_source_MT(source_name, i)
            n += s.fAASkippedParticles
    else:
        n = sim.get_source(source_name).fAASkippedParticles
    return n



import opengate as gate
import opengate_core as g4
import uproot
import numpy as np


class SourceInfoActor(g4.GateVActor, gate.ActorBase):
    """
    TODO
    """

    type_name = "SourceInfoActor"

    def __init__(self, name):
        g4.GateVActor.__init__(self, self.type_name)
        gate.ActorBase.__init__(self, name)
        # default actions
        self.actions = ["BeginOfRunAction", "EndOfRunAction", "BeginOfEventAction"]
        # parameters
        self.user_info.filename = None
        self.tree = None
        self.file = None
        # FIXME --> do it by batch
        self.positions = []

    def initialize(self):
        gate.ActorBase.initialize(self)
        if not self.user_info.filename:
            gate.fatal(
                f"Provide a filename to the actor {self.user_info.physics_list_name}"
            )
        # create the root tree
        self.file = uproot.recreate(self.user_info.filename)
        self.file[self.user_info.physics_list_name] = uproot.newtree(
            {
                "position_x": np.float64,
                "position_y": np.float64,
                "position_z": np.float64,
            }
        )
        self.tree = self.file[self.user_info.physics_list_name]
        print(self.tree)

    def BeginOfRunAction(self, run):
        print("Start run SourceInfoActor")

    def EndOfRunAction(self, run):
        print("End run SourceInfoActor")
        print(len(self.positions))
        self.positions = np.array(self.positions)
        self.tree.extend(
            {
                "position_x": self.positions[:, 0],
                "position_y": self.positions[:, 1],
                "position_z": self.positions[:, 2],
            }
        )

    def BeginOfEventAction(self, event):
        p = event.GetPrimaryVertex(0).GetPosition()
        # print('BeginOfEventAction')
        self.positions.append(gate.vec_g4_as_np(p))



import opengate as gate
from scipy.spatial.transform import Rotation

# colors (similar to the ones of Gate)
red = [1, 0, 0, 1]
blue = [0, 0, 1, 1]
green = [0, 1, 0, 1]
yellow = [0.9, 0.9, 0.3, 1]
gray = [0.5, 0.5, 0.5, 1]
white = [1, 1, 1, 0.8]


def create_material():
    gcm3 = gate.g4_units("g/cm3")
    gate.new_material_nb_atoms(f"ABS", 1.04 * gcm3, ["C", "H", "N"], [15, 17, 1])
    gate.new_material_weights(f"Copper", 8.920 * gcm3, "Cu")
    gate.new_material_nb_atoms(
        f"LYSO", 7.1 * gcm3, ["Lu", "Y", "Si", "O"], [18, 2, 10, 50]
    )
    gate.new_material_nb_atoms(f"LYSO_debug", 7.1 * gcm3, ["Lu", "O"], [1, 50])
    gate.new_material_weights(f"Lead", 11.16 * gcm3, ["Pb", "Sb"], [0.95, 0.05])
    gate.new_material_nb_atoms(f"Lexan", 1.2 * gcm3, ["C", "H", "O"], [15, 16, 2])
    gate.new_material_weights(f"CarbonFiber", 1.78 * gcm3, "C")


def add_pet(sim, name="pet", create_housing=True, create_mat=True):
    """
    Geometry of a PET Philips VEREOS
    Salvadori J, Labour J, Odille F, Marie PY, Badel JN, Imbert L, Sarrut D.
    Monte Carlo simulation of digital photon counting PET.
    EJNMMI Phys. 2020 Apr 25;7(1):23.
    doi: 10.1186/s40658-020-00288-w
    """

    # unit
    mm = gate.g4_units("mm")

    # define the materials (if needed)
    if create_mat:
        create_material()

    # ring volume
    pet = sim.add_volume("Tubs", name)
    pet.rmax = 500 * mm
    pet.rmin = 360 * mm
    pet.dz = 164 * mm / 2.0
    pet.color = gray
    pet.material = "G4_AIR"

    # ------------------------------------------
    # 18 modules
    #   of 4x5 stack
    #       of 4x4 die
    #           of 2x2 crystal
    # ------------------------------------------

    # Module (each module has 4x5 stacks)
    module = sim.add_volume("Box", f"{name}_module")
    module.mother = pet.name
    module.size = [19 * mm, 131.4 * mm, 164 * mm]
    module.translation = None
    module.rotation = None
    module.material = "ABS"
    module.color = blue
    le = gate.repeat_ring(module.name, 190, 18, [391.5 * mm, 0, 0], [0, 0, 1])
    module.repeat = le

    # Stack (each stack has 4x4 die)
    stack = sim.add_volume("Box", f"{name}_stack")
    stack.mother = module.name
    stack.size = [module.size[0], 32.6 * mm, 32.6 * mm]
    stack.material = "G4_AIR"
    stack.translation = None
    stack.rotation = None
    le = gate.repeat_array(stack.name, [1, 4, 5], [0, 32.85 * mm, 32.85 * mm])
    stack.repeat = le
    stack.color = green

    # Die (each die has 2x2 crystal)
    die = sim.add_volume("Box", f"{name}_die")
    die.mother = stack.name
    die.size = [module.size[0], 8 * mm, 8 * mm]
    die.material = "G4_AIR"
    die.translation = None
    die.rotation = None
    le = gate.repeat_array(die.name, [1, 4, 4], [0, 8 * mm, 8 * mm])
    die.repeat = le
    die.color = white

    # Crystal
    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = die.name
    crystal.size = [module.size[0], 4 * mm, 4 * mm]
    crystal.material = "LYSO"
    crystal.translation = None
    crystal.rotation = None
    le = gate.repeat_array(crystal.name, [1, 2, 2], [0, 4 * mm, 4 * mm])
    crystal.repeat = le
    crystal.color = red

    # ------------------------------------------
    # Housing
    # ------------------------------------------

    if not create_housing:
        return pet

    # SiPMs HOUSING
    housing = sim.add_volume("Box", f"{name}_housing")
    housing.mother = pet.name
    housing.size = [1 * mm, 131 * mm, 164 * mm]
    housing.translation = None  # [408 * mm, 0, 0]
    housing.rotation = None
    housing.material = "G4_AIR"
    housing.color = yellow
    le = gate.repeat_ring(module.name, 190, 18, [408 * mm, 0, 0], [0, 0, 1])
    housing.repeat = le

    # SiPMs UNITS
    sipms = sim.add_volume("Box", f"{name}_sipms")
    sipms.mother = housing.name

    sipms.size = [1 * mm, 32.6 * mm, 32.6 * mm]
    sipms.translation = None
    sipms.rotation = None
    sipms.material = "G4_AIR"
    sipms.color = green
    spacing = 32.8 * mm
    le = gate.repeat_array(sipms.name, [1, 4, 5], [0, spacing, spacing])
    sipms.repeat = le

    # cooling plate
    coolingplate = sim.add_volume("Box", f"{name}_coolingplate")
    coolingplate.mother = pet.name
    coolingplate.size = [30 * mm, 130.2 * mm, 164 * mm]
    coolingplate.translation = None
    coolingplate.rotation = None
    coolingplate.material = "Copper"
    coolingplate.color = blue
    le = gate.repeat_ring(module.name, 190, 18, [430 * mm, 0, 0], [0, 0, 1])
    coolingplate.repeat = le

    # ------------------------------------------
    # Shielding
    # ------------------------------------------

    # end shielding 1
    endshielding1 = sim.add_volume("Tubs", f"{name}_endshielding1")
    endshielding1.mother = "world"
    endshielding1.translation = [0, 0, 95 * mm]
    endshielding1.rmax = 410 * mm
    endshielding1.rmin = 362.5 * mm
    endshielding1.dz = 25 * mm / 2.0
    endshielding1.material = "Lead"
    endshielding1.color = yellow

    # end shielding 2
    endshielding2 = sim.add_volume("Tubs", f"{name}_endshielding2")
    endshielding2.mother = "world"
    endshielding2.translation = [0, 0, -95 * mm]
    endshielding2.rmax = 410 * mm
    endshielding2.rmin = 362.5 * mm
    endshielding2.dz = 25 * mm / 2.0
    endshielding2.material = "Lead"
    endshielding2.color = yellow

    # cover Lexan layer
    cover = sim.add_volume("Tubs", f"{name}_cover")
    cover.mother = "world"
    cover.translation = [0, 0, 0]
    cover.rmax = 355.5 * mm
    cover.rmin = 354 * mm
    cover.dz = 392 * mm / 2.0
    cover.material = "Lexan"
    cover.color = white

    return pet


def add_pet_debug(sim, name="pet", create_housing=True, create_mat=True):
    """
    FIXME DEBUG
    """

    # unit
    mm = gate.g4_units("mm")

    # define the materials (if needed)
    if create_mat:
        create_material()

    # ring volume
    pet = sim.add_volume("Tubs", name)
    pet.rmax = 500 * mm
    pet.rmin = 360 * mm
    pet.dz = 164 * mm / 2.0
    pet.color = gray
    pet.material = "G4_AIR"

    # ------------------------------------------
    # 18 modules
    #   of 4x5 stack
    #       of 4x4 die
    #           of 2x2 crystal
    # ------------------------------------------

    # Module (18 modules around 360 deg)
    module = sim.add_volume("Box", f"{name}_crystal")
    module.mother = pet.name
    module.size = [19 * mm, 131.4 * mm, 164 * mm]
    module.translation = None
    module.rotation = None
    module.material = "LYSO"
    module.color = blue
    le = gate.repeat_ring(module.name, 190, 18, [391.5 * mm, 0, 0], [0, 0, 1])
    module.repeat = le

    """# Stack (4x5 in a module)
    stack = sim.add_volume("Box", f"{name}_stack")
    stack.mother = module.name
    stack.size = [module.size[0], 32.6 * mm, 32.6 * mm]
    stack.material = "G4_AIR"
    stack.translation = None
    stack.rotation = None
    le = gate.repeat_array(stack.name, [1, 4, 5], [0, 32.85 * mm, 32.85 * mm])
    stack.repeat = le
    stack.color = green

    # Die (4x4 in a stack)
    die = sim.add_volume("Box", f"{name}_die")
    die.mother = stack.name
    die.size = [module.size[0], 8 * mm, 8 * mm]
    die.material = "G4_AIR"
    die.translation = None
    die.rotation = None
    le = gate.repeat_array(die.name, [1, 4, 4], [0, 8 * mm, 8 * mm])
    die.repeat = le
    die.color = white

    # Crystal (2x2 in a die)
    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = die.name
    crystal.size = [module.size[0], 4 * mm, 4 * mm]
    crystal.material = "LYSO"
    crystal.translation = None
    crystal.rotation = None
    le = gate.repeat_array(crystal.name, [1, 2, 2], [0, 4 * mm, 4 * mm])
    crystal.repeat = le
    crystal.color = red"""

    # ------------------------------------------
    # Housing
    # ------------------------------------------

    if not create_housing:
        return pet

    # SiPMs HOUSING
    housing = sim.add_volume("Box", f"{name}_housing")
    housing.mother = pet.name
    housing.size = [1 * mm, 131 * mm, 164 * mm]
    housing.translation = None  # [408 * mm, 0, 0]
    housing.rotation = None
    housing.material = "G4_AIR"
    housing.color = yellow
    le = gate.repeat_ring(module.name, 190, 18, [408 * mm, 0, 0], [0, 0, 1])
    housing.repeat = le

    # SiPMs UNITS
    sipms = sim.add_volume("Box", f"{name}_sipms")
    sipms.mother = housing.name

    sipms.size = [1 * mm, 32.6 * mm, 32.6 * mm]
    sipms.translation = None
    sipms.rotation = None
    sipms.material = "G4_AIR"
    sipms.color = green
    spacing = 32.8 * mm
    le = gate.repeat_array(sipms.name, [1, 4, 5], [0, spacing, spacing])
    sipms.repeat = le

    # cooling plate
    coolingplate = sim.add_volume("Box", f"{name}_coolingplate")
    coolingplate.mother = pet.name
    coolingplate.size = [30 * mm, 130.2 * mm, 164 * mm]
    coolingplate.translation = None
    coolingplate.rotation = None
    coolingplate.material = "Copper"
    coolingplate.color = blue
    le = gate.repeat_ring(module.name, 190, 18, [430 * mm, 0, 0], [0, 0, 1])
    coolingplate.repeat = le

    # ------------------------------------------
    # Shielding
    # ------------------------------------------

    # end shielding 1
    endshielding1 = sim.add_volume("Tubs", f"{name}_endshielding1")
    endshielding1.mother = "world"
    endshielding1.translation = [0, 0, 95 * mm]
    endshielding1.rmax = 410 * mm
    endshielding1.rmin = 362.5 * mm
    endshielding1.dz = 25 * mm / 2.0
    endshielding1.material = "Lead"
    endshielding1.color = yellow

    # end shielding 2
    endshielding2 = sim.add_volume("Tubs", f"{name}_endshielding2")
    endshielding2.mother = "world"
    endshielding2.translation = [0, 0, -95 * mm]
    endshielding2.rmax = 410 * mm
    endshielding2.rmin = 362.5 * mm
    endshielding2.dz = 25 * mm / 2.0
    endshielding2.material = "Lead"
    endshielding2.color = yellow

    # cover Lexan layer
    cover = sim.add_volume("Tubs", f"{name}_cover")
    cover.mother = "world"
    cover.translation = [0, 0, 0]
    cover.rmax = 355.5 * mm
    cover.rmin = 354 * mm
    cover.dz = 392 * mm / 2.0
    cover.material = "Lexan"
    cover.color = white

    return pet


def add_table(sim, name="pet"):
    """
    Add a patient table
    """

    # unit
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")
    deg = gate.g4_units("deg")

    # main bed
    bed = sim.add_volume("Tubs", f"{name}_bed")
    bed.mother = "world"
    bed.rmax = 439 * mm
    bed.rmin = 406 * mm
    bed.dz = 200 * cm / 2.0
    bed.sphi = 0 * deg
    bed.dphi = 70 * deg
    bed.translation = [0, 25 * cm, 0]
    bed.rotation = Rotation.from_euler("z", -125, degrees=True).as_matrix()
    bed.material = "CarbonFiber"
    bed.color = white

    # interior of the bed
    bedin = sim.add_volume("Tubs", f"{name}_bedin")
    bedin.mother = bed.name
    bedin.rmax = 436.5 * mm
    bedin.rmin = 408.5 * mm
    bedin.dz = 200 * cm / 2.0
    bedin.sphi = 0 * deg
    bedin.dphi = 69 * deg
    bedin.translation = [0, 0, 0]
    bedin.rotation = Rotation.from_euler("z", 0.5, degrees=True).as_matrix()
    bedin.material = "G4_AIR"
    bedin.color = red

    return bed



import opengate_core.opengate_core

import opengate as gate
from opengate import log


class ActorManager:
    """
    Manage all the actors in the simulation
    """

    def __init__(self, simulation):
        self.simulation = simulation
        self.user_info_actors = {}
        self.actors = {}
        self.action_manager = None

    def __str__(self):
        v = [v.name for v in self.user_info_actors.values()]
        s = f'{" ".join(v)} ({len(self.user_info_actors)})'
        return s

    def __del__(self):
        pass

    def dump(self):
        n = len(self.user_info_actors)
        s = f"Number of Actors: {n}"
        for actor in self.user_info_actors.values():
            if n > 1:
                a = "\n" + "-" * 20
            else:
                a = ""
            a += f"\n {actor}"
            s += gate.indent(2, a)
        return s

    def get_actor(self, name):
        if name not in self.actors:
            gate.fatal(
                f"The actor {name} is not in the current "
                f"list of actors: {self.actors}"
            )
        return self.actors[name]

    def get_actor_user_info(self, name):
        if name not in self.user_info_actors:
            gate.fatal(
                f"The actor {name} is not in the current "
                f"list of actors: {self.user_info_actors}"
            )
        return self.user_info_actors[name]

    def add_actor(self, actor_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_actors, name)
        # build it
        a = gate.UserInfo("Actor", actor_type, name)
        # append to the list
        self.user_info_actors[name] = a
        # return the info
        return a

    def create_actors(self, action_manager):
        self.action_manager = action_manager
        for ui in self.user_info_actors.values():
            actor = gate.new_element(ui, self.simulation)
            log.debug(f"Actor: initialize [{ui.type_name}] {ui.name}")
            actor.initialize()
            self.actors[ui.name] = actor

    def initialize(self):
        # consider the priority value of the actors
        l = [
            {"name": l, "priority": self.actors[l].user_info.priority}
            for l in self.actors
        ]
        sorted_actors = sorted(l, key=lambda d: d["priority"])
        # for actor in self.actors.values():
        for ac in sorted_actors:
            actor = self.actors[ac["name"]]
            log.debug(
                f"Actor: initialize [{actor.user_info.type_name}] {actor.user_info.name}"
            )
            self.register_all_actions(actor)
            # warning : the step actions will be registered by register_sensitive_detectors
            # called by ConstructSDandField

    def register_all_actions(self, actor):
        # Run
        for ra in self.action_manager.g4_RunAction:
            ra.RegisterActor(actor)
        # Event
        for ea in self.action_manager.g4_EventAction:
            ea.RegisterActor(actor)
        # Track
        for ta in self.action_manager.g4_TrackingAction:
            ta.RegisterActor(actor)
        # initialization
        actor.ActorInitialize()

    def register_sensitive_detector_propagate(self, actor, vol):
        lvs = opengate_core.opengate_core.G4LogicalVolumeStore.GetInstance()
        lv = lvs.GetVolume(vol, False)
        for i in range(lv.GetNoDaughters()):
            da = lv.GetDaughter(i)
            self.register_sensitive_detector_to_childs(actor, da.GetLogicalVolume())
            self.register_sensitive_detector_propagate(
                actor, da.GetLogicalVolume().GetName()
            )

    def register_sensitive_detectors(self):
        for actor in self.actors.values():
            if not "SteppingAction" in actor.fActions:
                continue
            # Step: only enabled if attachTo a given volume.
            # Propagated to all child and sub-child
            tree = self.simulation.volume_manager.volumes_tree
            mothers = actor.user_info.mother
            if isinstance(mothers, str):
                # make a list with one single element
                mothers = [mothers]
            # add SD for all mothers
            for vol in mothers:
                if vol not in tree:
                    s = (
                        f"Cannot attach the actor {actor.user_info.name} "
                        f"because the volume {vol} does not exists"
                    )
                    gate.fatal(s)
                # Propagate the Geant4 Sensitive Detector to all children
                lv = self.simulation.volume_manager.volumes[vol].g4_logical_volume
                self.register_sensitive_detector_to_childs(actor, lv)

                # FIXME find all daughters ???
                # self.register_sensitive_detector_propagate(actor, vol)

    def register_sensitive_detector_to_childs(self, actor, lv):
        log.debug(
            f'Actor: "{actor.user_info.name}" '
            f'(attached to "{actor.user_info.mother}") '
            f'set to volume "{lv.GetName()}"'
        )
        actor.RegisterSD(lv)
        n = lv.GetNoDaughters()
        for i in range(n):
            child = lv.GetDaughter(i).GetLogicalVolume()
            self.register_sensitive_detector_to_childs(actor, child)

    def start_simulation(self):
        for actor in self.actors.values():
            actor.StartSimulationAction()

    def stop_simulation(self):
        for actor in self.actors.values():
            actor.EndSimulationAction()



import opengate as gate


class FilterBase(gate.UserElement):
    """
    Store user information about a filter
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # no user properties for all filters (maybe later)

    def __init__(self, user_info):
        # type_name MUST be defined in class that inherit from FilterBase
        super().__init__(user_info)

    def __del__(self):
        pass

    def __str__(self):
        s = f"str FilterBase {self.user_info.name} of type {self.user_info.type_name}"
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_base import *

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main description
spect = create_spect_simu(sim, paths, 1)
proj = test_add_proj(sim, paths)

# rotate spect
cm = gate.g4_units("cm")
psd = 6.11 * cm
p = [0, 0, -(20 * cm + psd)]
spect.translation, spect.rotation = gate.get_transform_orbiting(p, "y", -15)

sim.initialize()
sim.start()

# check
test_spect_proj(sim, paths, proj)



import opengate as gate
import opengate_core as g4


class PolyhedraVolume(gate.VolumeBase):
    type_name = "Polyhedra"

    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        cm = gate.g4_units("cm")
        deg = gate.g4_units("deg")
        user_info.phi_start = 0 * deg
        user_info.phi_total = 360 * deg
        user_info.num_side = 6
        user_info.num_zplanes = 2
        h = 5 * cm
        user_info.zplane = [-h / 2, h - h / 2]
        user_info.radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        user_info.radius_outer = [
            0.15 * cm,
            0.15 * cm,
            0.15 * cm,
            0.15 * cm,
            0.15 * cm,
            0.15 * cm,
        ]

    def build_solid(self):
        u = self.user_info
        return g4.G4Polyhedra(
            u.name,
            u.phi_start,
            u.phi_total,
            u.num_side,
            u.num_zplanes,
            u.zplane,
            u.radius_inner,
            u.radius_outer,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test008_dose_actor")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.number_of_threads = 2

#  change world size
m = gate.g4_units("m")
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a simple fake volume to test hierarchy
# translation and rotation like in the Gate macro
fake = sim.add_volume("Box", "fake")
cm = gate.g4_units("cm")
fake.size = [40 * cm, 40 * cm, 40 * cm]
fake.translation = [1 * cm, 2 * cm, 3 * cm]
fake.rotation = Rotation.from_euler("x", 10, degrees=True).as_matrix()
fake.material = "G4_AIR"
fake.color = [1, 0, 1, 1]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.mother = "fake"
waterbox.size = [10 * cm, 10 * cm, 10 * cm]
waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
waterbox.rotation = Rotation.from_euler("y", 20, degrees=True).as_matrix()
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]

# default source for tests
source = sim.add_source("Generic", "mysource")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.energy.mono = 150 * MeV
nm = gate.g4_units("nm")
source.particle = "proton"
source.position.radius = 1 * nm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
# source.activity = 2e5 / sim.user_info.number_of_threads * Bq
source.activity = 5e4 / sim.user_info.number_of_threads * Bq

"""
It needs at least around 2e5 particles for the multithread to be faster than mono thread
"""

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test012-edep.mhd"
dose.mother = "waterbox"
dose.size = [99, 99, 99]
mm = gate.g4_units("mm")
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.translation = [2 * mm, 3 * mm, -2 * mm]

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# create G4 objects
sim.initialize()

# explicit check overlap (already performed during initialize)
sim.check_volumes_overlap(verbose=True)

# print info
print(sim.dump_volumes())

# verbose
# sim.apply_g4_command('/tracking/verbose 0')
sim.apply_g4_command("/run/verbose 2")
# sim.apply_g4_command("/event/verbose 2")
# sim.apply_g4_command("/tracking/verbose 1")

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("Stats")
print(stat)

dose = sim.get_actor("dose")
print(dose)

# tests
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
# change the number of run to the number of threads
stats_ref.counts.run_count = sim.user_info.number_of_threads
is_ok = gate.assert_stats(stat, stats_ref, 0.05)

is_ok = (
    gate.assert_images(
        paths.gate_output / "output-Edep.mhd",
        paths.output / "test012-edep.mhd",
        stat,
        tolerance=45,
    )
    and is_ok
)
gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
import json
import itk
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
ui.check_volumes_overlap = False

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
MeV = gate.g4_units("MeV")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
kBq = Bq * 1000

# world
sim.world.size = [3 * m, 3 * m, 3 * m]

# add a first iec phantom (analytical)
iec1 = gate_iec.add_phantom(sim, "iec1")
iec1.translation = [40 * cm, 0 * cm, 0 * cm]
# rotation should have no effect
iec1.rotation = Rotation.from_euler("y", 33, degrees=True).as_matrix()

# to highlight the position, we change some volume with high density lead
v = sim.get_volume_user_info("iec1_sphere_37mm")
v.material = "G4_LEAD_OXIDE"
v = sim.get_volume_user_info("iec1_center_cylinder_hole")
v.material = "G4_LEAD_OXIDE"

# add a second iec phantom (voxelized)
iec2 = sim.add_volume("Image", "iec2")
iec2.image = paths.output_ref / "test032_iec.mhd"
iec2.material = "G4_AIR"
iec2.translation = [-40 * cm, 0 * cm, 0 * cm]
iec2.dump_label_image = paths.output / "test032_iec_label.mhd"
labels = json.loads(open(paths.output_ref / "test032_labels.json").read())
iec2.voxel_materials = []
for l in labels:
    mat = "IEC_PLASTIC"
    if "capillary" in l:
        mat = "G4_WATER"
    if "cylinder_hole" in l:
        mat = "G4_LUNG_ICRP"
    if "interior" in l:
        mat = "G4_WATER"
    if "sphere" in l:
        mat = "G4_WATER"
    if "world" in l:
        mat = "G4_AIR"
    if "shell" in l:
        mat = "IEC_PLASTIC"
    if "sphere_37mm" in l:
        mat = "G4_LEAD_OXIDE"
    if "center_cylinder_hole" in l:
        mat = "G4_LEAD_OXIDE"
    m = [labels[l], labels[l] + 1, mat]
    iec2.voxel_materials.append(m)

pMin, pMax = gate.get_volume_bounding_limits(sim, "iec1")
print(f"pMin and pMax of iec1", pMin, pMax)

# the origin of iec1 is different from the origin of iec2
# we create fake images to be able to convert from
# the image coordinate space to iec1 or iec2
# Coordinate system of iec1 is pMin (the extend)
# Coordinate system of iec2 is the center of the image bounding box
img = itk.imread(str(iec2.image))
fake1 = gate.create_image_like(img)
pMin = gate.vec_g4_as_np(pMin)
fake1.SetOrigin(pMin)

fake2 = gate.create_image_like(img)
info = gate.get_info_from_image(fake2)
origin = -info.size * info.spacing / 2.0 + info.spacing / 2.0
fake2.SetOrigin(origin)

# sources
activity = 10 * kBq
for i in range(1, 3):
    source = sim.add_source("Generic", f"src{i}")
    source.mother = f"iec{i}"
    source.energy.mono = 100 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 10 * mm
    # WARNING the center of the volume is different in the image (iec2)
    # and in the analytical phantom (iec1)
    p = [31 * mm, 33 * mm, 36 * mm]
    if i == 1:
        p = gate.transform_images_point(p, img, fake1)
    else:
        p = gate.transform_images_point(p, img, fake2)
    source.position.translation = p
    source.activity = activity
    source.direction.type = "iso"

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "stats")
stats.track_types_flag = True

# add dose actor
for i in range(1, 3):
    dose = sim.add_actor("DoseActor", f"dose{i}")
    dose.output = paths.output / f"test032_iec{i}_edep.mhd"
    dose.mother = f"iec{i}"
    dose.size = [100, 100, 100]
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    # translate the iec1 to have the exact same dose origin
    # (only needed to perform the assert_image test)
    if i == 1:
        dose.translation = [0 * mm, 35 * mm, 0 * mm]
    dose.img_coord_system = True

# initialize & start
sim.initialize()
sim.start()

# stats
stats = sim.get_actor("stats")
print(stats)

# compare edep map
is_ok = gate.assert_images(
    paths.output / "test032_iec1_edep.mhd",
    paths.output / "test032_iec2_edep.mhd",
    stats,
    tolerance=79,
    axis="x",
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import numpy as np
import opengate_core as g4

v = g4.G4ThreeVector(2)
print("v =", v)
assert v == g4.G4ThreeVector(2, 0, 0)

print("v.x =", v.x)
v.x = 1
v.y = 5
v.z = 4
print("v=", v)
assert v == g4.G4ThreeVector(1, 5, 4)

print(f"v mag = {v.mag()}")
np.testing.assert_almost_equal(v.mag(), 6.48074069840786)

w = g4.G4ThreeVector(v)
print(f"w = {w}")
assert v == w
w.x = 6
w.y = 5
w.z = 4
print(f"w = {w}")
assert w == g4.G4ThreeVector(6, 5, 4)
assert v == g4.G4ThreeVector(1, 5, 4)

print(f"2*v = {2 * v}")
assert 2 * v == g4.G4ThreeVector(2, 10, 8)

print(f"v.w = {v * w}")
assert v * w == 47.0

v -= w
print(f"v -= w -> {v}")
assert v == g4.G4ThreeVector(-5, 0, 0)

v[0] = 666
print(v)
try:
    v[40] = 12
    print(v)
    print(v[12])
    print(v[40])
except:
    print("Index error ok")

print("tol", g4.G4ThreeVector.getTolerance())

# overloaded
# help(v.isNear)
print("is near", v.isNear(w, 20))
assert v.isNear(w, 20) == True

print("is near", v.isNear(w))
assert v.isNear(w) == False

gate.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import itk
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
print(ui)

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq

#  change world size
world = sim.world
world.size = [1.5 * m, 1 * m, 1 * m]

# fake box
b = sim.add_volume("Box", "fake1")
b.size = [36 * cm, 36 * cm, 36 * cm]
b.translation = [25 * cm, 0, 0]
r = Rotation.from_euler("y", -25, degrees=True)
r = r * Rotation.from_euler("x", -35, degrees=True)
b.rotation = r.as_matrix()
b = sim.add_volume("Box", "fake2")
b.mother = "fake1"
b.size = [35 * cm, 35 * cm, 35 * cm]

# CT image #1
ct_odd = sim.add_volume("Image", "ct_odd")
ct_odd.image = paths.data / "10x10x10.mhd"
ct_odd.mother = "fake2"
ct_odd.voxel_materials = [[0, 10, "G4_WATER"]]
ct_odd.translation = [-2 * cm, 0, 0]
r = Rotation.from_euler("y", 25, degrees=True)
r = r * Rotation.from_euler("x", 35, degrees=True)
ct_odd.rotation = r.as_matrix()

# CT image #2
ct_even = sim.add_volume("Image", "ct_even")
ct_even.image = paths.data / "11x11x11.mhd"
ct_even.voxel_materials = [[0, 10, "G4_WATER"]]
ct_even.voxel_materials = ct_odd.voxel_materials
ct_even.translation = [-25 * cm, 0, 0]

# source from sphere
"""
    WARNING : if the source is a point and is centered with odd image, the source
    is at the intersection of 3 planes (8 voxels): then, lot of "navigation warning"
    from G4 occur. Not really clear why.
    So we move the source a bit.
"""
source = sim.add_source("Generic", "s_odd")
source.particle = "alpha"
source.activity = 1000 * Bq / ui.number_of_threads
source.direction.type = "iso"
source.mother = "ct_odd"
source.position.translation = [10 * mm, 10 * mm, 10 * mm]
source.energy.mono = 1 * MeV

# source from sphere
source = sim.add_source("Generic", "s_even")
source.particle = "alpha"
source.activity = 1000 * Bq / ui.number_of_threads
source.direction.type = "iso"
source.mother = "ct_even"
source.position.translation = [0 * mm, 0 * mm, 0 * mm]
source.energy.mono = 1 * MeV

# source from spect
source = sim.add_source("Voxels", "vox")
source.mother = "ct_even"
source.particle = "alpha"
source.activity = 1000 * Bq / ui.number_of_threads
source.image = paths.data / "five_pixels.mha"
source.direction.type = "iso"
source.position.translation = [0 * mm, 0 * mm, 0 * mm]
source.position.translation = gate.get_translation_between_images_center(
    str(ct_even.image), str(source.image)
)
source.energy.mono = 1 * MeV

# cuts
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"
p.enable_decay = False
sim.set_cut("world", "all", 1 * mm)

# add dose actor
dose1 = sim.add_actor("DoseActor", "dose1")
dose1.output = paths.output / "test021-odd-edep.mhd"
dose1.mother = "ct_odd"
img_info = gate.read_image_info(str(ct_odd.image))
dose1.size = img_info.size
dose1.spacing = img_info.spacing
dose1.img_coord_system = True

# add dose actor
dose2 = sim.add_actor("DoseActor", "dose2")
dose2.output = paths.output / "test021-even-edep.mhd"
dose2.mother = "ct_even"
img_info = gate.read_image_info(str(ct_even.image))
dose2.size = img_info.size
dose2.spacing = img_info.spacing
dose2.translation = source.position.translation
dose2.img_coord_system = True

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# create G4 objects
sim.initialize()

# verbose
sim.apply_g4_command("/tracking/verbose 0")

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("Stats")
# stat.write('output_ref/stat021_ref.txt')

# test pixels in dose #1
d_odd = itk.imread(str(dose1.output))
s = itk.array_view_from_image(d_odd).sum()
v = d_odd.GetPixel([5, 5, 5])
diff = (s - v) / s
tol = 0.01
is_ok = diff < tol
diff *= 100
gate.print_test(is_ok, f"Image #1 (odd): {v:.2f} {s:.2f} -> {diff:.2f}%")

# test pixels in dose #1
d_even = itk.imread(str(dose2.output))
s = itk.array_view_from_image(d_even).sum()
v0 = d_even.GetPixel([5, 5, 5])
v1 = d_even.GetPixel([1, 5, 5])
v2 = d_even.GetPixel([1, 2, 5])
v3 = d_even.GetPixel([5, 2, 5])
v4 = d_even.GetPixel([6, 2, 5])
tol = 0.15
ss = v0 + v1 + v2 + v3 + v4


def t(s, v):
    diff = abs(s - v) / s
    b = diff < tol
    p = diff * 100.0
    gate.print_test(b, f"Image #2 (even) {s:.2f} vs {v:.2f}  -> {p:.2f}%")
    return b


is_ok = t(s, ss) and is_ok
is_ok = t(1.80, v0) and is_ok
is_ok = t(0.8, v1) and is_ok
is_ok = t(0.8, v2) and is_ok
is_ok = t(0.8, v3) and is_ok
is_ok = t(0.8, v4) and is_ok

stats_ref = gate.read_stat_file(paths.output_ref / "stat021_ref.txt")
stats_ref.counts.run_count = ui.number_of_threads
is_ok = gate.assert_stats(stat, stats_ref, 0.05) and is_ok

gate.test_ok(is_ok)



import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read()

all_packages = find_packages()
selected_packages = []
for p in all_packages:
    if "opengate_core" not in p:
        selected_packages.append(p)

setuptools.setup(
    name="opengate",
    version=version,
    author="Opengate collaboration",
    author_email="david.sarrut@creatis.insa-lyon.fr",
    description="Simulation for Medical Physics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenGATE/opengate",
    packages=selected_packages,
    python_requires=">=3.7",
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        "opengate_core==" + version,
        "gatetools",
        "tqdm",
        "colored",
        "click",
        "python-box",
        "anytree",
        "numpy",
        "itk",
        "uproot",
        "sphinx",
        "scipy",
        "sphinx_pdj_theme",
        "matplotlib",
        "myst-parser",
        "GitPython",
        "colorlog",
        "sphinx_copybutton",
        "autoapi",
        "sphinx-autoapi",
    ],
    scripts=[
        "opengate/bin/opengate_tests",
        "opengate/bin/opengate_tests_wip",
        "opengate/bin/opengate_tests_utils",
        "opengate/bin/opengate_info",
        "opengate/bin/opengate_user_info",
        "opengate/bin/dose_rate",
        "opengate/bin/split_spect_projections",
        "opengate/bin/voxelize_iec_phantom",
    ],
)



import opengate as gate
from .GenericSource import *
import opengate_core as g4
import numpy as np


class PBSource(GenericSource):
    """
    PB source
    """

    type_name = "PB"

    @staticmethod
    def set_default_user_info(user_info):

        GenericSource.set_default_user_info(user_info)
        # additional parameters
        # position
        user_info.position = (
            Box()
        )  # Box() resets the object to blank. All param set for position before this line are cancelled
        user_info.position.type = "disc"
        user_info.position.size = [0, 0, 0]
        user_info.position.translation = [0, 0, 0]
        user_info.position.rotation = Rotation.identity().as_matrix()
        user_info.position.confine = None
        # direction
        user_info.direction.partPhSp_x = [
            0,
            0,
            0,
            0,
        ]  # sigma, theta, epsilon, conv (0: divergent, 1:convergent)
        user_info.direction.partPhSp_y = [0, 0, 0, 0]

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GatePBSource()

    def __init__(self, user_info):
        super().__init__(user_info)

    def set_transform_from_user_info(self):
        pass
        # get source image information

    #        src_info = gate.read_image_info(str(self.user_info.image))
    #        # get pointer to SPSVoxelPosDistribution
    #        pg = self.g4_source.GetSPSVoxelPosDistribution()
    #        # update cpp image info (no need to allocate)
    #        gate.update_image_py_to_cpp(self.image, pg.cpp_edep_image, False)
    #        # set spacing
    #        pg.cpp_edep_image.set_spacing(src_info.spacing)
    #        # set origin (half size + translation and half pixel shift)
    #        c = -src_info.size / 2.0 * src_info.spacing + self.user_info.position.translation + src_info.spacing / 2.0
    #        pg.cpp_edep_image.set_origin(c)

    def cumulative_distribution_functions(self):
        """
        Compute the Cumulative Distribution Function of the image
        Composed of: CDF_Z = 1D, CDF_Y = 2D, CDF_X = 3D
        """
        pass

    #        cdf_x, cdf_y, cdf_z = gate.compute_image_3D_CDF(self.image)
    #
    #        # set CDF to the position generator
    #        pg = self.g4_source.GetSPSVoxelPosDistribution()
    #        pg.SetCumulativeDistributionFunction(cdf_z, cdf_y, cdf_x)

    def initialize(self, run_timing_intervals):
        # read source image

        # compute position
        self.set_transform_from_user_info()

        # create Cumulative Distribution Function
        # self.cumulative_distribution_functions()

        # initialize standard options (particle energy, etc)
        # we temporarily set the position attribute to reuse
        # the GenericSource verification
        GenericSource.initialize(self, run_timing_intervals)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__)

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.check_volumes_overlap = False

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("s")

#  change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add two water boxes
wb1 = sim.add_volume("Box", "waterbox1")
wb1.size = [20 * cm, 20 * cm, 20 * cm]
wb1.translation = [-20 * cm, 0, 0]
wb2 = sim.add_volume("Box", "waterbox2")
wb2.size = [20 * cm, 20 * cm, 20 * cm]
wb2.translation = [20 * cm, 0, 0]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
p.enable_decay = True
sim.set_cut("world", "all", 0.1 * mm)
# p.energy_range_min = 250 * eV
# p.energy_range_max = 15 * MeV

# sources info
activity = 10 * Bq
hl = 6586.26 * sec  # 109.771 minutes

# source ion
ion_src = sim.add_source("Generic", "ion_source")
ion_src.mother = wb1.name
ion_src.particle = "ion 9 18"  # F18
ion_src.position.type = "sphere"
ion_src.position.radius = 10 * mm
ion_src.direction.type = "iso"
ion_src.energy.type = "mono"
ion_src.energy.mono = 0
ion_src.half_life = hl
ion_src.activity = activity

# source e+
beta_src = sim.add_source("Generic", "beta+_source")
beta_src.mother = wb2.name
beta_src.particle = "e+"
beta_src.position.type = "sphere"
beta_src.position.radius = 10 * mm
beta_src.energy.type = "F18"
beta_src.direction.type = "iso"
beta_src.half_life = hl
total_yield = gate.get_rad_yield("F18")
beta_src.activity = activity * total_yield

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# phsp
phsp1 = sim.add_actor("PhaseSpaceActor", "phsp_ion")
phsp1.mother = wb1.name
phsp1.attributes = [
    "KineticEnergy",
    "LocalTime",
    "GlobalTime",
    "TrackProperTime",
    "TimeFromBeginOfEvent",
    # 'TrackVertexKineticEnergy', 'EventKineticEnergy'
]
phsp1.output = paths.output / "test013_decay_ion.root"
f = sim.add_filter("ParticleFilter", "f")
f.particle = "e+"
phsp1.filters.append(f)

phsp2 = sim.add_actor("PhaseSpaceActor", "phsp_beta")
phsp2.mother = wb2.name
phsp2.attributes = phsp1.attributes
phsp2.output = paths.output / "test013_decay_beta_plus.root"
phsp2.filters.append(f)

# long run
sim.run_timing_intervals = [[0, 109 * 60 * sec]]

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results
stats = sim.get_actor("Stats")
print(stats)

print()
keys1 = phsp1.attributes
keys2 = keys1
scalings = [1] * len(keys1)
scalings[2] = 1e-12  # GlobalTime
tols = [0.008] * len(keys1)
tols[2] = 0.04  # GlobalTime
print(keys2, scalings, tols)
print(phsp1.output)
print(phsp2.output)
print()
is_ok = gate.compare_root3(
    phsp1.output,
    phsp2.output,
    "phsp_ion",
    "phsp_beta",
    keys1,
    keys2,
    tols,
    scalings,
    scalings,
    paths.output / "test013_decay.png",
)

gate.test_ok(is_ok)



from .BoxVolume import *
from .SphereVolume import *
from .TrapVolume import *
from .ImageVolume import *
from .TubsVolume import *
from .ConsVolume import *
from .PolyhedraVolume import *
from .TrdVolume import *
from .BooleanVolume import *
from .RepeatParametrisedVolume import *

volume_type_names = {
    BoxVolume,
    SphereVolume,
    TrapVolume,
    ImageVolume,
    TubsVolume,
    PolyhedraVolume,
    ConsVolume,
    TrdVolume,
    BooleanVolume,
    RepeatParametrisedVolume,
}
volume_builders = gate.make_builders(volume_type_names)

# G4Tubs G4CutTubs G4Cons G4Para G4Trd
# G4Torus (G4Orb not needed) G4Tet
# G4EllipticalTube G4Ellipsoid G4EllipticalCone
# G4Paraboloid G4Hype
# specific: G4Polycone G4GenericPolycone Polyhedra
# G4ExtrudedSolid G4TwistedBox G4TwistedTrap G4TwistedTrd G4GenericTrap
# G4TwistedTubs

"""
http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html#constructed-solid-geometry-csg-solids
"""


def box_add_size(box, thickness):
    box.size = [x + thickness for x in box.size]


def cons_add_size(cons, thickness):
    cons.rmax1 += thickness / 2
    cons.rmax2 += thickness / 2
    cons.dz += thickness


def get_volume_bounding_limits(simulation, volume_name):
    """
    Return the min and max 3D points of the bounding box of the given volume
    """
    v = simulation.get_volume_user_info(volume_name)
    s = simulation.get_solid_info(v)
    pMin = s.bounding_limits[0]
    pMax = s.bounding_limits[1]
    return pMin, pMax


def get_volume_bounding_box_size(simulation, volume_name):
    """
    Return the size of the bounding box of the given volume
    """
    pMin, pMax = get_volume_bounding_limits(simulation, volume_name)
    return [pMax[0] - pMin[0], pMax[1] - pMin[1], pMax[2] - pMin[2]]


def translate_point_to_volume(simulation, volume, top, x):
    """

    Consider the point x in the current volume and return the coordinate of x in the top volume
    (that must be an ancestor).
    Translation only, do not consider rotation.
    """
    while volume.name != top:
        x += volume.translation
        volume = simulation.get_volume_user_info(volume.mother)
    return x


# correspondence element names <> symbol
elements_name_symbol = {
    "Hydrogen": "H",
    "Carbon": "C",
    "Nitrogen": "N",
    "Oxygen": "O",
    "Sodium": "Na",
    "Magnesium": "Mg",
    "Phosphor": "P",
    "Sulfur": "S",
    "Chlorine": "Cl",
    "Argon": "Ar",
    "Potassium": "K",
    "Calcium": "Ca",
    "Titanium": "Ti",
    "Copper": "Cu",
    "Zinc": "Zn",
    "Silver": "Ag",
    "Tin": "Sn",
}



import opengate as gate
import json


class UserInfo:
    """
    A simple dict that contains the list of user parameters.
    Note that the dict is a Box, allowing simpler access to the keys with a dot
    (rather than brackets)

    The default elements are set with set_default_user_info according to
    the class found thanks to element_type and type_name
    """

    def __init__(self, element_type, type_name, name=None):
        # set the element and the type (it will be checked in get_element_class later)
        # element_type is Volume, Source or Actor
        self.element_type = element_type
        self.type_name = type_name
        # set the name
        self._name = name
        # set the default parameters and values
        cl = gate.get_element_class(element_type, type_name)
        cl.set_default_user_info(self)

    @property
    def name(self):
        # make 'name' a property make it read_only.
        # user cannot change the name of the object once it is declared
        return self._name

    def __str__(self):
        s = f"{self.element_type} {self.name} : {self.__dict__}"
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# global log level
# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

#  change world size
m = gate.g4_units("m")
cm = gate.g4_units("cm")
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]

# add a iec phantom
iec_phantom = gate_iec.add_phantom_old(sim)
iec_phantom.translation = [0 * cm, 3.5 * cm, 0 * cm]

# simple source
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source = sim.add_source("Generic", "g")
source.particle = "gamma"
source.energy.mono = 0.1 * MeV
source.direction.type = "iso"
source.activity = 50000 * Bq

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "stats")
stats.track_types_flag = True

# run timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 1 * sec]]

# initialize & start
sim.initialize()
sim.start()

# print results at the end
stats = sim.get_actor("stats")
print(stats)
stats.write(pathFile / ".." / "output" / "stats_test015_iec_phantom_1.txt")

# check
stats_ref = gate.SimulationStatisticsActor()
c = stats_ref.counts
c.run_count = 1
c.event_count = 49997
c.track_count = 53027
c.step_count = 468582
# stats_ref.pps = 2150
sec = gate.g4_units("second")
c.duration = c.event_count / 19441.5 * sec
print("-" * 80)
is_ok = gate.assert_stats(stats, stats_ref, 0.06)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_acc_angle_base import *
import itk
import numpy as np

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main description
spect, proj = create_spect_simu(sim, paths, number_of_threads=2, activity_kBq=1000)

ui = sim.user_info
# ui.force_multithread_mode = True
ui.running_verbose_level = 0  # 50 for event
ui.random_engine = "MixMaxRng"
ui.visu = False
print(ui)

# rotate spect
cm = gate.g4_units("cm")
psd = 6.11 * cm
p = [0, 0, -(15 * cm + psd)]
spect.translation, spect.rotation = gate.get_transform_orbiting(p, "y", 15)
print("translation", spect.translation)

sim.initialize()
# sim.apply_g4_command("/run/eventModulo 5000 1")
sim.start()

# check
# test_spect_proj(sim, paths, proj)

gate.warning("Compare acceptance angle skipped particles")
stats = sim.get_actor("Stats")
reference_ratio = 691518 / 2998895  # (23%)
b1 = gate.get_source_skipped_particles(sim, "beam1")
b2 = gate.get_source_skipped_particles(sim, "beam2")
b3 = gate.get_source_skipped_particles(sim, "beam3")
tol = 0.3
r1 = b1 / stats.counts.event_count
is_ok = (r1 - reference_ratio) / reference_ratio < tol
gate.print_test(
    is_ok,
    f"Skipped particles b1 = {b1} {r1 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
)

r2 = b2 / stats.counts.event_count
is_ok = (r2 - reference_ratio) / reference_ratio < tol
gate.print_test(
    is_ok,
    f"Skipped particles b2 = {b2} {r2 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
)

r3 = b3 / stats.counts.event_count
is_ok = (r3 - reference_ratio) / reference_ratio < tol
gate.print_test(
    is_ok,
    f"Skipped particles b3 = {b3} {r3 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
)

# stat
gate.warning("Compare stats")
print(stats)
stats_ref = gate.read_stat_file(paths.gate_output / "stat4.txt")
print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
stats.counts.run_count = 1  # force to 1
print(
    f"Number of steps was {stats.counts.step_count}, force to the same value (because of angle acceptance). "
)
stats.counts.step_count = stats_ref.counts.step_count  # force to id
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07) and is_ok

# read image and force change the offset to be similar to old Gate
gate.warning("Compare projection image")
img = itk.imread(str(paths.output / "proj028_colli.mhd"))
spacing = np.array(proj.spacing)
origin = spacing / 2.0
origin[2] = 0.5
spacing[2] = 1
img.SetSpacing(spacing)
img.SetOrigin(origin)
itk.imwrite(img, str(paths.output / "proj028_colli_offset.mhd"))
# There are not enough event to make a proper comparison, so the tol is very high
is_ok = (
    gate.assert_images(
        paths.gate_output / "projection4.mhd",
        paths.output / "proj028_colli_offset.mhd",
        stats,
        tolerance=85,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)

gate.test_ok(is_ok)



from box import Box
import opengate as gate
import opengate_core as g4
import sys


def create_phys_list(physics):
    # set physics list
    factory = g4.G4PhysListFactory()
    phys_list = factory.GetReferencePhysList(physics.physics_list_name)
    # FIXME check if exist
    # FIXME EM only PL to create
    return phys_list


def set_cuts(physics, g4_PhysList):
    # set cuts
    # g4_PhysList.DumpList()
    # g4_PhysList.DumpCutValuesTable(1)
    print("default cut value", g4_PhysList.GetDefaultCutValue())
    pct = g4.G4ProductionCutsTable.GetProductionCutsTable()
    # print('pct', pct)
    eV = gate.g4_units("eV")
    GeV = gate.g4_units("GeV")
    pct.SetEnergyRange(250 * eV, 100 * GeV)
    # print('default cut value', g4_PhysList.GetDefaultCutValue())


"""
    The following functions are a mechanism to automatically create G4VModularPhysicsList
    from simple G4VPhysicsConstructor.

    Current available lists:
    G4EmStandardPhysics_option1
    G4EmStandardPhysics_option2
    G4EmStandardPhysics_option3
    G4EmStandardPhysics_option4

--> Example of dynamically created class:
class G4EmStandardPhysics_option1(g4.G4VModularPhysicsList):
    def __init__(self):
        g4.G4VModularPhysicsList.__init__(self)
        self.p = g4.G4EmStandardPhysics_option1(1)
        self.RegisterPhysics(self.p)
"""

# Names of the PL that can be created dynamically
available_additional_physics_lists = [
    "G4EmStandardPhysics_option1",
    "G4EmStandardPhysics_option2",
    "G4EmStandardPhysics_option3",
    "G4EmStandardPhysics_option4",
    "G4EmStandardPhysicsGS",
    "G4EmLowEPPhysics",
    "G4EmLivermorePhysics",
    "G4EmLivermorePolarizedPhysics",
    "G4EmPenelopePhysics",
    "G4EmDNAPhysics",
    "G4OpticalPhysics",
]


def create_modular_physics_list_class(pl_class):
    """
    Create a class (not on object!) which:
    - inherit from g4.G4VModularPhysicsList
    - register a single G4 PhysicList (G4VPhysicsConstructor)
    - is named with the name of this PhysicList
    """
    # get the name of the G4 PhysicList
    name = pl_class.__name__
    # create the class with constructor
    the_class = type(
        name,
        (g4.G4VModularPhysicsList,),
        {
            "pl_class": pl_class,
            "__init__": modular_physics_list_constructor,
            "__del__": modular_physics_list_destructor,
        },
    )
    return the_class


def modular_physics_list_constructor(self):
    """
    Constructor of the above, dynamically created class.
    - call the constructor of the super class (G4VModularPhysicsList)
    - Create and register the physic list (G4VPhysicsConstructor)
    """
    g4.G4VModularPhysicsList.__init__(self)
    self.p = self.pl_class(1)
    self.RegisterPhysics(self.p)


def modular_physics_list_destructor(self):
    """
    For debug
    """
    pass


def create_modular_physics_list(pl_name):
    """
    Dynamically create a class with the given PhysicList
    Only possible if the class exist in g4
    """
    # Retrieve the G4VPhysicsConstructor class
    try:
        a = getattr(sys.modules["opengate_core"], pl_name)
    except:
        s = f"Cannot find the class {pl_name} in opengate_core"
        gate.fatal(s)
    # Create the class
    b = gate.create_modular_physics_list_class(a)
    # Create the object
    return b()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
from scipy.spatial.transform import Rotation
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
MeV = gate.g4_units("MeV")
KeV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
kBq = Bq * 1000
cm3 = gate.g4_units("cm3")
BqmL = Bq / cm3

#  change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a iec phantom
iec_phantom = gate_iec.add_phantom(sim)
iec_phantom.translation = [-5 * cm, -1 * cm, 2 * cm]
iec_phantom.rotation = Rotation.from_euler("z", 33, degrees=True).as_matrix()

# simple source
ac = 1000 * BqmL
sources = gate_iec.add_spheres_sources(
    sim, "iec", "iec_source", [10, 13, 17, 22, 28, 37], [ac, ac, ac, ac, ac, ac]
)
for s in sources:
    s.particle = "e-"
    s.energy.type = "mono"
    s.energy.mono = 1 * MeV

# Central source in "lung" compartment
name = iec_phantom.name
bg1 = sim.add_source("Generic", "bg1")
bg1.mother = f"{name}_center_cylinder_hole"
v = sim.get_volume_user_info(bg1.mother)
s = sim.get_solid_info(v)
bg_volume = s.cubic_volume / cm3
print(f"Volume of {bg1.mother} {bg_volume} cm3")
bg1.position.type = "box"
bg1.position.size = gate.get_volume_bounding_box_size(sim, bg1.mother)
bg1.position.confine = bg1.mother
bg1.particle = "e-"
bg1.energy.type = "mono"
bg1.energy.mono = 1 * MeV
bg1.activity = ac * s.cubic_volume / 3  # ratio with spheres

# background source
# (I checked that source if confine only on mother, not including daughter volumes)
bg2 = sim.add_source("Generic", "bg2")
bg2.mother = f"{name}_interior"
v = sim.get_volume_user_info(bg2.mother)
s = sim.get_solid_info(v)
bg_volume = s.cubic_volume / cm3
print(f"Volume of {bg2.mother} {bg_volume} cm3")
bg2.position.type = "box"
bg2.position.size = gate.get_volume_bounding_box_size(sim, bg2.mother)
bg2.position.confine = bg2.mother
bg2.particle = "e-"
bg2.energy.type = "mono"
bg2.energy.mono = 1 * MeV
bg2.activity = ac * s.cubic_volume / 10  # ratio with spheres

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "stats")
stats.track_types_flag = True

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = pathFile / ".." / "output" / "test015_confine.mhd"
# dose.output = 'output_ref/test015_confine.mhd'
dose.mother = "iec"
dose.size = [200, 200, 200]
dose.spacing = [2 * mm, 2 * mm, 2 * mm]

# initialize & start
sim.initialize()
sim.start()

# Only for reference stats:
stats = sim.get_actor("stats")
stats.write(pathFile / ".." / "output" / "test015_confine_stats.txt")
# stats.write('output_ref/test015_confine_stats.txt')

# check
stats_ref = gate.read_stat_file(
    pathFile / ".." / "data" / "output_ref" / "test015_confine_stats.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, 0.03)
is_ok = is_ok and gate.assert_images(
    pathFile / ".." / "data" / "output_ref" / "test015_confine.mhd",
    pathFile / ".." / "output" / "test015_confine.mhd",
    stats,
    tolerance=78,
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__, "gate_test043_garf")

m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
nm = gate.g4_units("nm")
km = gate.g4_units("km")
gcm3 = gate.g4_units("g/cm3")
MeV = gate.g4_units("MeV")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq


def sim_set_world(sim):
    # world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "G4_AIR"

    return world


def sim_add_detector_plane(sim, spect_name, distance, plane_name="detPlane"):
    # detector input plane
    detector_plane = sim.add_volume("Box", plane_name)
    detector_plane.mother = spect_name
    detector_plane.size = [57.6 * cm, 44.6 * cm, 1 * nm]
    detector_plane.translation = [0, 0, distance]
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [0, 1, 0, 1]

    return detector_plane


def sim_phys(sim):
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    sim.set_cut("world", "all", 1 * km)


def sim_source_test(sim, activity):
    w, e = gate.get_rad_energy_spectrum("Tc99m")

    # first sphere
    s1 = sim.add_source("Generic", "s1")
    s1.particle = "gamma"
    s1.activity = activity * 0.0001
    s1.position.type = "sphere"
    s1.position.radius = 15 * mm
    s1.position.translation = [0, 0, 0]
    s1.direction.type = "momentum"
    s1.direction.momentum = [0, 0, -1]
    s1.energy.type = "spectrum"
    s1.energy.spectrum_energy = e
    s1.energy.spectrum_weight = w

    # second sphere
    s2 = sim.add_source("Generic", "s2")
    s2.particle = "gamma"
    s2.activity = activity * 2
    s2.position.type = "sphere"
    s2.position.radius = 15 * mm
    s2.position.translation = [15 * cm, 0, 0]
    s2.direction.type = "iso"
    s2.energy.type = "spectrum"
    s2.energy.spectrum_energy = e
    s2.energy.spectrum_weight = w

    # third sphere
    s3 = sim.add_source("Generic", "s3")
    s3.particle = "gamma"
    s3.activity = activity
    s3.position.type = "sphere"
    s3.position.radius = 28 * mm
    s3.position.translation = [-10 * cm, 5 * cm, 0]
    s3.direction.type = "iso"
    s3.energy.type = "spectrum"
    s3.energy.spectrum_energy = e
    s3.energy.spectrum_weight = w



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import bisect
import numpy as np
import itk
import opengate as gate
import opengate_core as g4


class VoxelizedSourcePDFSampler:
    """
    This is an alternative to GateSPSVoxelsPosDistribution (c++)
    It is needed because the cond voxel source is used on python side.

    There are two versions, version 2 is much slower (do not use)
    """

    def __init__(self, itk_image, version=1):
        self.image = itk_image
        self.version = version
        # get image in np array
        self.imga = itk.array_view_from_image(itk_image)
        imga = self.imga

        # image sizes
        lx = self.imga.shape[0]
        ly = self.imga.shape[1]
        lz = self.imga.shape[2]
        m = lx * ly * lz

        # normalized pdf
        pdf = imga.ravel(order="F")
        self.pdf = pdf / pdf.sum()

        # create grid of indices
        [x_grid, y_grid, z_grid] = np.meshgrid(
            np.arange(lx), np.arange(ly), np.arange(lz), indexing="ij"
        )

        # list of indices
        self.xi, self.yi, self.zi = (
            x_grid.ravel(order="F"),
            y_grid.ravel(order="F"),
            z_grid.ravel(order="F"),
        )

        # 1D indices
        self.linear_indices = np.arange(int(m))

        if version == 2:
            self.init_cdf()

        if version == 3:
            self.init_cdf()

        # ------------------------------------------
        # Below = NOT WORKING, DO NOT USE
        # create grid of indices in physical space
        """img_info = gate.get_info_from_image(self.image)
        sp = img_info.spacing
        hsp = sp / 2
        s = img_info.size * sp
        [px_grid, py_grid, pz_grid] = np.meshgrid(
            np.linspace(hsp[0], s[0] - hsp[0], lx),
            np.linspace(hsp[1], s[1] - hsp[1], lx),
            np.linspace(hsp[2], s[2] - hsp[2], lz),
            indexing='ij',
        )
        # list of indices
        self.pxi, self.pyi, self.pzi = (
            px_grid.ravel(order='F'),
            py_grid.ravel(order='F'),
            pz_grid.ravel(order='F'),
        )
        """
        # ------------------------------------------

    def init_cdf(self):
        self.cdf_x, self.cdf_y, self.cdf_z = gate.compute_image_3D_CDF(self.image)
        self.cdf_x = np.array(self.cdf_x)
        self.cdf_y = np.array(self.cdf_y)
        self.cdf_z = np.array(self.cdf_z)
        self.cdf_init = True

    def searchsorted2d(self, a, b):
        # https://stackoverflow.com/questions/56471109/how-to-vectorize-with-numpy-searchsorted-in-a-2d-array
        # Inputs : a is (m,n) 2D array and b is (m,) 1D array.
        # Finds np.searchsorted(a[i], b[i])) in a vectorized way by
        # scaling/offsetting both inputs and then using searchsorted

        # Get scaling offset and then scale inputs
        s = np.r_[0, (np.maximum(a.max(1) - a.min(1) + 1, b) + 1).cumsum()[:-1]]
        a_scaled = (a + s[:, None]).ravel()
        b_scaled = b + s

        # Use searchsorted on scaled ones and then subtract offsets
        return np.searchsorted(a_scaled, b_scaled) - np.arange(len(s)) * a.shape[1]

    def sample_indices_slower(self, n, rs=np.random):
        """
        This version seems slower than the other version with np random choice
        """
        # Z (here, search sorted is faster than bisect+loop)
        uz = rs.uniform(0, 1, size=n)
        # i = [bisect.bisect_left(self.cdf_z, uz[t], lo=0, hi=lz) for t in range(n)]
        i = np.searchsorted(self.cdf_z, uz, side="left")

        # Y, knowing Z
        # https://stackoverflow.com/questions/56471109/how-to-vectorize-with-numpy-searchsorted-in-a-2d-array
        ly = self.imga.shape[1]
        uy = rs.uniform(0, 1, size=n)
        j = [bisect.bisect_left(self.cdf_y[i[t]], uy[t], lo=0, hi=ly) for t in range(n)]

        # (here search sorted is not faster, we keep bisect)
        # slower:
        # cdfyi = np.take(self.cdf_y, i, axis=0)
        # j = self.searchsorted2d(cdfyi, uy)

        # X
        lx = self.imga.shape[2]
        ux = rs.uniform(0, 1, size=n)
        k = [
            bisect.bisect_left(self.cdf_x[i[t]][j[t]], ux[t], lo=0, hi=lx)
            for t in range(n)
        ]

        return i, j, k

    def samples_g4(self, n):
        # to compare with cpp version
        sps = g4.GateSPSVoxelsPosDistribution()
        sps.SetCumulativeDistributionFunction(self.cdf_z, self.cdf_y, self.cdf_x)
        p = np.array([sps.VGenerateOneDebug() for a in range(n)])
        return p[:, 2], p[:, 1], p[:, 0]

    def sample_indices(self, n, rs=np.random):
        indices = rs.choice(self.linear_indices, size=n, replace=True, p=self.pdf)
        i = self.xi[indices]
        j = self.yi[indices]
        k = self.zi[indices]
        return i, j, k

    def sample_indices_phys(self, n, rs=np.random):
        # TODO (not used yet)
        indices = rs.choice(self.linear_indices, size=n, replace=True, p=self.pdf)
        i = self.pxi[indices]
        j = self.pyi[indices]
        k = self.pzi[indices]
        return i, j, k



import logging
import colorlog
import sys

# https://github.com/borntyping/python-colorlog
# The available color names are black,
# red, green, yellow, blue, purple, cyan and white.

# main format
formatter = colorlog.ColoredFormatter(
    "%(log_color)s%(log_color)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={"NONE": "cyan", "DEBUG": "cyan", "INFO": "green"},
    secondary_log_colors={},
    style="%",
)

# set output message to standard output
handler = colorlog.StreamHandler(sys.stdout)

# install default handler format
handler.setFormatter(formatter)

# get main log object
log = colorlog.getLogger(__name__)
log.addHandler(handler)

# default log level
log.setLevel(logging.INFO)

# shorter for level
NONE = 0
DEBUG = logging.DEBUG
INFO = logging.INFO
RUN = 20
EVENT = 50



import opengate as gate
import opengate_core as g4


class MotionVolumeActor(g4.GateMotionVolumeActor, gate.ActorBase):
    """
    Every run, move a volume according to the given translations and rotations.
    """

    type_name = "MotionVolumeActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.translations = []
        user_info.rotations = []
        user_info.priority = 10

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        # check rotations and translation
        u = user_info
        if len(u.translations) != len(u.rotations):
            gate.fatal(
                f"Error, translations and rotations must have the same length, while it is"
                f" {len(u.translations)} and {len(u.rotations)}"
            )
        g4.GateMotionVolumeActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)

    def __del__(self):
        pass

    def __str__(self):
        s = f"MotionVolumeActor {self.user_info.name}"
        return s

    def initialize(self):
        super().initialize()
        # check translations and rotations
        rt = self.simulation.run_timing_intervals
        ui = self.user_info
        if len(ui.translations) != len(rt):
            gate.fatal(
                f"Error in actor {ui}. "
                f"Translations must be the same length than the number of runs. "
                f"While it is {len(ui.translations)} instead of {len(rt)}"
            )
        if len(ui.rotations) != len(rt):
            gate.fatal(
                f"Error in actor {ui}. "
                f"Rotations must be the same length than the number of runs. "
                f"While it is {len(ui.rotations)} instead of {len(rt)}"
            )



import opengate as gate
import opengate_core as g4


class PhaseSpaceActor(g4.GatePhaseSpaceActor, gate.ActorBase):
    """
    Similar to HitsCollectionActor : store a list of hits.
    However only the first hit of given event is stored here.
    """

    type_name = "PhaseSpaceActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        # options
        user_info.attributes = []
        user_info.output = f"{user_info.name}.root"
        user_info.store_absorbed_event = False
        user_info.debug = False

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GatePhaseSpaceActor.__init__(self, user_info.__dict__)
        self.fStepFillNames = user_info.attributes  # this is a copy

    def __del__(self):
        pass

    def __str__(self):
        s = f"PhaseSpaceActor {self.user_info.name}"
        return s

    # not needed, only if need to do something from python
    def StartSimulationAction(self):
        g4.GatePhaseSpaceActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GatePhaseSpaceActor.EndSimulationAction(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test043_garf_helpers import *
import opengate.contrib.spect_ge_nm670 as gate_spect

paths = gate.get_default_test_paths(__file__, "gate_test043_garf")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.number_of_threads = 1
ui.visu = False
ui.random_seed = 123654

# activity
activity = 1e6 * Bq / ui.number_of_threads

# world size
sim_set_world(sim)

# spect head
spect = gate_spect.add_ge_nm67_spect_head(
    sim, "spect", collimator_type="lehr", debug=ui.visu
)
crystal_name = f"{spect.name}_crystal"

# detector input plane
pos, crystal_dist, psd = gate_spect.get_plane_position_and_distance_to_crystal("lehr")
pos += 1 * nm  # to avoid overlap
print(f"plane position     {pos / mm} mm")
print(f"crystal distance   {crystal_dist / mm} mm")
detPlane = sim_add_detector_plane(sim, spect.name, pos)

# physics
sim_phys(sim)

# source
s1 = sim.add_source("Generic", "s1")
s1.particle = "gamma"
s1.activity = activity
s1.position.type = "disc"
s1.position.radius = 57.6 * cm / 4  # FIXME why ???
s1.position.translation = [0, 0, 12 * cm]
s1.direction.type = "iso"
s1.energy.type = "range"
s1.energy.min_energy = 0.01 * MeV
s1.energy.max_energy = 0.154 * MeV
s1.direction.acceptance_angle.volumes = [detPlane.name]
s1.direction.acceptance_angle.intersection_flag = True

# digitizer
channels = [
    {"name": f"scatter_{spect.name}", "min": 114 * keV, "max": 126 * keV},
    {"name": f"peak140_{spect.name}", "min": 126 * keV, "max": 154 * keV},
]
cc = gate_spect.add_digitizer_energy_windows(sim, crystal_name, channels)

# arf actor for building the training dataset
arf = sim.add_actor("ARFTrainingDatasetActor", "ARF (training)")
arf.mother = detPlane.name
arf.output = paths.output / "test043_arf_training_dataset.root"
arf.energy_windows_actor = cc.name
arf.russian_roulette = 100

dpz = detPlane.translation[2]
print(f"Position of the detector plane {dpz} mm")

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True
s.output = str(arf.output).replace(".root", "_stats.txt")

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("stats")
print(stat)
skip = gate.get_source_skipped_particles(sim, "s1")
print(f"Nb of skip particles {skip}  {(skip / stat.counts.event_count) * 100:.2f}%")

# ----------------------------------------------------------------------------------------------------------------
gate.warning("Compare stats")
stats_ref = gate.read_stat_file(paths.output_ref / s.output)
is_ok = gate.assert_stats(stat, stats_ref, 0.01)

gate.warning("Compare root")
checked_keys = [
    {"k1": "E", "k2": "E", "tol": 0.002, "scaling": 1},
    {"k1": "Theta", "k2": "Theta", "tol": 2, "scaling": 1},
    {"k1": "Phi", "k2": "Phi", "tol": 1.5, "scaling": 1},
    {"k1": "window", "k2": "window", "tol": 0.006, "scaling": 1},
]
is_ok = (
    gate.compare_root2(
        paths.output_ref / "test043_arf_training_dataset.root",
        arf.output,
        "ARF (training)",
        "ARF (training)",
        checked_keys,
        paths.output / "test043_training_dataset.png",
        n_tol=14,
    )
    and is_ok
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test043_garf_helpers import *
import itk
import opengate.contrib.spect_ge_nm670 as gate_spect

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.number_of_threads = 1
ui.visu = False
ui.random_seed = "auto"

# activity
activity = 1e6 * Bq / ui.number_of_threads

# add a material database
sim.add_material_database(paths.gate_data / "GateMaterials.db")

# init world
sim_set_world(sim)

# fake spect head
head = gate_spect.add_ge_nm67_fake_spect_head(sim, "spect")
head.translation = [0, 0, -15 * cm]

# detector input plane (+ 1nm to avoid overlap)
pos, crystal_dist, psd = gate_spect.get_plane_position_and_distance_to_crystal("lehr")
pos += 1 * nm
print(f"plane position     {pos / mm} mm")
print(f"crystal distance   {crystal_dist / mm} mm")
detPlane = sim_add_detector_plane(sim, head.name, pos)

# physics
sim_phys(sim)

# sources
sim_source_test(sim, activity)

# arf actor
arf = sim.add_actor("ARFActor", "arf")
arf.mother = detPlane.name
arf.output = paths.output / "test043_projection_garf.mhd"
arf.batch_size = 2e5
arf.image_size = [128, 128]
arf.image_spacing = [4.41806 * mm, 4.41806 * mm]
arf.verbose_batch = True
arf.distance_to_crystal = crystal_dist  # 74.625 * mm
arf.distance_to_crystal = 74.625 * mm
arf.pth_filename = paths.gate_data / "pth" / "arf_Tc99m_v3.pth"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("stats")
print(stat)

# print info
print("")
arf = sim.get_actor("arf")
img = arf.output_image
# set the first channel to the same channel (spectrum) than the analog
img[0, :] = img[1, :] + img[2, :]
print(f"Number of batch: {arf.batch_nb}")
print(f"Number of detected particles: {arf.detected_particles}")
filename1 = str(arf.user_info.output).replace(".mhd", "_0.mhd")
itk.imwrite(img, filename1)

# high stat
filename2 = str(arf.user_info.output).replace(".mhd", "_hs.mhd")
scale = 4e8 * Bq / activity
print(f"Scaling ref = 4e8, activity = {activity}, scale = {scale}")
img2 = gate.scale_itk_image(img, scale)
itk.imwrite(img2, filename2)

# ----------------------------------------------------------------------------------------------------------------
# tests
print()
gate.warning("Tests stats file")
stats_ref = gate.read_stat_file(paths.gate_output / "stats_analog.txt")
# dont compare steps of course
stats_ref.counts.step_count = stat.counts.step_count
is_ok = gate.assert_stats(stat, stats_ref, 0.01)

print()
gate.warning("Compare image to analog")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test043_projection_analog.mhd",
        filename1,
        stat,
        tolerance=100,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)

print()
gate.warning("Compare image to analog high statistics")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test043_projection_analog_high_stat.mhd",
        filename2,
        stat,
        tolerance=52,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)

print()
gate.warning("profile compare : ")
p = paths.output_ref
print(
    f'garf_compare_image_profile {p / "test043_projection_analog.mhd"} {filename1} -w 3'
)
print(
    f'garf_compare_image_profile {p / "test043_projection_analog.mhd"} {filename1} -w 3 -s 75'
)
print(
    f'garf_compare_image_profile {p / "test043_projection_analog_high_stat.mhd"} {filename2} -w 3'
)
print(
    f'garf_compare_image_profile {p / "test043_projection_analog_high_stat.mhd"} {filename2} -w 3 -s 75'
)

gate.delete_run_manager_if_needed(sim)
gate.test_ok(is_ok)
